import os, sys, shutil
from pathlib import Path
from textwrap import dedent
import hashlib
import logging


# 3rd Party
import pytest

if (p := os.path.abspath("../")) not in sys.path:
    sys.path.insert(0, p)


from dfb import rclonerc
from dfb.rclonerc import RC, rcpathjoin, rcpathsplit


def rmdir(path):
    try:
        shutil.rmtree(path)
    except OSError:
        pass


# do all tests in one  since it can then use the same rc
def test_main():
    # if True:
    testpath = Path("testdirs/rcmain")
    rmdir(testpath)
    testpath.mkdir(exist_ok=True, parents=False)

    config = dedent(
        f"""\
    [src]
    type = alias
    remote = {str(testpath / 'A')}

    [dst]
    type = alias
    remote = {str(testpath / 'B')}
    """
    )

    (testpath / "config.cfg").write_text(config)

    rc_log = logging.getLogger("dfb.rclonerc-rc-server")
    rc_log.setLevel(logging.DEBUG)
    stream_handler = logging.StreamHandler(stream=sys.stderr)
    rc_log.addHandler(stream_handler)

    rc = RC(
        serve_flags=["--config", str(testpath / "config.cfg"), "-vv"],
        rclone_env={"RCLONE_PASSWORD_COMMAND": RC.DELENV},
    )

    # test config settings
    # Test the config setting in flags
    assert set(rc.call("config/listremotes").get("remotes", [])) == {"src", "dst"}

    # Write and read some files
    rc.write("src:file1.txt", b"file1")
    rc.write("src:sub/file2.txt", b"file two")

    assert rc.read("src:file1.txt").strip() == b"file1"
    assert rc.read("src:sub/file2.txt").strip() == b"file two"

    # copy/move/delete
    rc.copyfile(src="src:file1.txt", dst=("dst:subb", "file1.txt"))
    rc.movefile(
        src=("src:", "sub/file2.txt"),
        dst="dst:file2.txt",
        _config={
            "NoCheckDest": True,
            "metadata": True,
        },
    )
    rc.delete("src:file1.txt")

    # verify
    assert not rc.list("src:", only="files")  # should be empty from the above

    # Test some of the listing. Verification comes later
    rc.list("dst:", only="dirs")
    rc.list("dst:", only="files")
    rc.list("dst:", only="files", epoch_time=True)

    assert {f["Path"] for f in rc.list("dst:", only="files")} == {
        "file2.txt",
        "subb/file1.txt",
    }

    assert {f["Path"] for f in rc.list("dst:", only="files", filters="- *2.txt")} == {
        "subb/file1.txt"
    }

    ## Test paths

    test_cases = [
        ("single-file.ext", ("./", "single-file.ext")),
        ("local/file.ext", ("local", "file.ext")),
        ("remote:file.ext", ("remote:", "file.ext")),
        ("remote:sub/file.ext", ("remote:", "sub/file.ext")),
        ("remote:/sub/file.ext", ("remote:", "/sub/file.ext")),
        (":http:sub/file.ext", (":http:", "sub/file.ext")),
        (
            ":http,url='https://example.com':path/to/dir",
            (":http,url='https://example.com':", "path/to/dir"),
        ),
        (
            ":http,url='https://example.com':path/t'o/dir/with'quote",
            (":http,url='https://example.com':", "path/t'o/dir/with'quote"),
        ),
    ]

    for full, gold in test_cases:
        split = rcpathsplit(full)
        assert split == gold
        assert rcpathjoin(*split).removeprefix("./") == full

    assert rcpathjoin("a", "b") == "a/b"
    assert rcpathjoin("a:", "b") == "a:b"
    assert rcpathjoin("a:", "/b") == "a:/b"  # Note that these are unlike os.path.join
    assert rcpathjoin("a", "/b") == "a/b"

    ## listing and stat and agreement on them
    # This is the same as stat on 'dst:file2.txt'
    _stat_from_list = lambda **p: [
        a for a in rc.list("dst:", only="files", **p) if a["Path"] == "file2.txt"
    ][0]

    assert _stat_from_list() == rc.stat("dst:file2.txt")

    assert _stat_from_list(hashes=True) == (
        stat := rc.stat("dst:file2.txt", hashes=True)
    )
    assert "md5" in stat["Hashes"]

    assert _stat_from_list(
        hashes=True, hashtypes=("crc32", "sha1"), epoch_time=True
    ) == (
        stat := rc.stat(
            "dst:file2.txt", hashes=True, hashtypes=("crc32", "sha1"), epoch_time=True
        )
    )
    assert set(stat["Hashes"]) == {"crc32", "sha1"}
    assert isinstance(stat["ModTime"], (float, int))

    ## noops

    rc.call_async_and_wait("rc/noop")
    rc.call("rc/noop")
    rc.call("rc/noopauth")

    try:
        rc.call("rc/error")
        assert False
    except rclonerc.RcloneError:
        pass

    ## File Objects. Test by comparing with a regular one
    assert (
        rc.read("dst:file2.txt")
        == rc.open("dst:file2.txt").read()
        == rc.open("dst:file2.txt", "rt").read().encode()  # Text mode
        == (testpath / "B" / "file2.txt").read_bytes()
    )

    testfile = testpath / "A" / "random.bin"
    with testfile.open("wb") as fp:
        mb1 = os.urandom(1024 * 1024)
        for _ in range(20):
            fp.write(mb1)

    h0 = hashlib.sha1()
    h1 = hashlib.sha1()

    with (
        testfile.open("rb") as fp0,
        rc.open("src:random.bin", buffer_size=6 * 1024 * 1024) as fp1,
    ):
        h0.update(fp0.read(5013))
        h1.update(fp1.read(5013))

        h0.update(fp0.read(1024 * 1024 * 7))
        h1.update(fp1.read(1024 * 1024 * 7))

        fp0.seek(0)
        fp1.seek(0)

        h0.update(fp0.read(100))
        h1.update(fp1.read(100))

        fp0.seek(-10, 1)
        fp1.seek(-10, 1)

        h0.update(fp0.read(100))
        h1.update(fp1.read(100))

        fp0.seek(-105, 2)
        fp1.seek(-105, 2)

        h0.update(fp0.read(100))
        h1.update(fp1.read(100))

        h0.update(fp0.read())
        h1.update(fp1.read())

        fp0.seek(0)
        fp1.seek(0)

        h0.update(fp0.read())
        h1.update(fp1.read())

        assert rcpathsplit(fp1) == ("src:", "random.bin")

        # print(f"{h0.hexdigest() = } == {h1.hexdigest() = }")
        assert h0.digest() == h1.digest()

    rc.check()
    rc.start(check=True)
    rc.stop()


if __name__ == "__main__":
    test_main()

    print("=" * 50)
    print(" PASS ".center(50, "="))
    print("=" * 50)



"""
CLI Listing-type functions
"""

import os, sys
import shutil
import json
import operator
import logging
import shlex
from textwrap import dedent

from . import LOCK, time2all
from .dstdb import DFBDST
from .utils import tabulate, human_readable_bytes, head_tail_table, smart_open
from .timestamps import timestamp_parser
from .rclonerc import rcpathjoin

logger = logging.getLogger(__name__)

STRFTIME_FMT = "%Y-%m-%dT%H:%M:%S"


def snapshot(config):
    args = config.cliconfig
    dstdb = DFBDST(config)

    rows = dstdb.snapshot(
        path=args.path,
        before=args.before,
        after=args.after,
        export=args.export,  # below is ignored if export.
        remove_delete=args.deleted == 0,
        delete_only=args.deleted > 1,
        add_query="ORDER BY LOWER(apath)",
    )
    rows = (dstdb.fullrow2dict(row) for row in rows)

    if args.output:
        parent, name = os.path.split(args.output)
        swap = os.path.join(parent, f".swap.{name}")
        with smart_open(swap, "wt") as fp:
            for row in rows:
                json.dump(row, fp)
                fp.write("\n")
            fp.flush()
        shutil.move(swap, args.output)
    else:
        for row in rows:
            print(json.dumps(row, ensure_ascii=False))
        print("", end="", flush=True)


def tree(config):
    # del -- Handled in snapshot
    # del del -- Handled in snapshot
    # depth -- Done with added condition

    args = config.cliconfig
    dstdb = DFBDST(config)
    groupselect = dedent(
        """
        * 
        -- LENGTH(apath) - LENGTH(REPLACE(apath, '/', '')) AS depth -- not used
        """
    )
    path = args.path.removesuffix("/").removeprefix("./")

    conditions = []
    # Depth filtering is not done here because it means that directories aren't listed.
    # This is much more performant but is less ideal because you won't see subdirs even if
    # files exist path the depth

    # if args.max_depth > 0:
    #    d0 = len(path.split("/")) if path else 0
    #    conditions.append(["depth <= :depth", {"depth": args.max_depth - 1 + d0}])

    rows = dstdb.snapshot(
        path=path,
        before=args.before,
        after=args.after,
        groupselect=groupselect,
        conditions=conditions,
        remove_delete=args.deleted == 0,
        delete_only=args.deleted > 1,
        add_query="ORDER BY LOWER(apath)",
    )

    rows = [dstdb.fullrow2dict(row) for row in rows]

    treedict = {}
    for row in rows:
        dpath = os.path.relpath(row["apath"], path).removeprefix("./")

        parts = dpath.split("/")
        current_node = treedict
        for part in parts[:-1]:
            if part not in current_node:
                current_node[part] = {}
            current_node = current_node[part]

        # 'row' is a dict so wrap it in a tuple for later type check
        current_node[parts[-1]] = (row,)

    def _print_tree(treedict, indent="", depth=1):
        """
        Print the nested dictionary as a tree structure with line characters.
        """
        if args.max_depth > 0 and depth > args.max_depth:
            return

        # Get the keys of the dictionary
        items = list(treedict.items())

        # Iterate over each key
        for i, (key, val) in enumerate(items):
            # Determine the prefix characters based on the position in the list
            if i == len(items) - 1:
                prefix = "└── "
                next_indent = indent + "    "
            else:
                prefix = "├── "
                next_indent = indent + "│   "

            # Print the current key with the appropriate prefix
            isdir = isinstance(val, dict)
            if isdir:
                rep = f"{key}/"
            else:
                row = val[0]  # from the tuple
                rep = key
                if row["size"] < 0:
                    rep = f"{rep} (DEL)"
            print(indent + prefix + rep, flush=True)

            # If the value associated with the key is a dictionary, recursively print its contents
            if isdir:
                _print_tree(val, next_indent, depth=depth + 1)

    print(f"{path}/", flush=True)
    _print_tree(treedict)


def ls(config):
    args = config.cliconfig
    dstdb = DFBDST(config)

    subdirs, files = dstdb.ls(
        subdir=args.path,
        before=args.before,
        after=args.after,
        remove_delete=args.deleted == 0,
        delete_only=args.deleted > 1,
        recursive=args.recursive,
    )

    subdirs = list(subdirs)

    # default args.list_only option
    if args.list_only is None:
        args.list_only = "files" if args.recursive else "both"

    if args.list_only == "dirs":
        files[:] = []
    elif args.list_only == "files":
        subdirs[:] = []
    # elif args.list_only in {None, "both"}":
    #    pass # Not needed but feels more complete
    # argparse will block other options

    items = subdirs + files
    items.sort(key=lambda i: i if isinstance(i, str) else i["apath"])

    # Build a table
    table = [["versions", "total_size", "size", "ModTime", "Timestamp", "path"]]

    for item in items:
        if isinstance(item, str):  # subdir
            if (sub := os.path.relpath(item, args.path)) == ".":
                continue

            item = item if args.full_path else sub
            table.append(["", "", "", "", "", f"{item.removesuffix('/')}/"])
            continue

        versions = str(item["versions"])

        mtime = item.get("mtime", None)
        if not mtime:
            mtime = ""
        else:
            mtime = (
                timestamp_parser(mtime, aware=True)
                .astimezone()
                .strftime(f"{STRFTIME_FMT}")
            )

        ts = item["timestamp"]
        ts = timestamp_parser(ts)
        if args.timestamp_local:
            ts = ts.astimezone().strftime(f"{STRFTIME_FMT}%z")
        else:
            ts = ts.strftime(f"{STRFTIME_FMT}Z")

        path = item["apath"]
        if args.rpath:  # If it's a reference, we'd prefer ref_rpath if not deleted
            if item["isref"] and args.rpath == 1 and item["size"] >= 0:
                path = item["ref_rpath"]
            else:
                path = item["rpath"]
        path = path if args.full_path else os.path.relpath(path, args.path)

        if args.human:
            size = "{:0.2f} {}".format(*human_readable_bytes(item["size"]))
            tot_size = "{:0.2f} {}".format(*human_readable_bytes(item["tot_size"]))
        else:
            size = str(item["size"])
            tot_size = str(item["tot_size"])

        if item["size"] < 0:
            path = f"{path} (DEL)"
            size = "D"
        elif item["isref"]:
            size = f"{size}R"
        table.append([versions, tot_size, size, mtime, ts, path])

    if args.long == 0:
        keep = ["path"]
    elif args.long == 1:
        keep = ["size", "ModTime", "path"]
    else:
        keep = ["versions", "total_size", "size", "ModTime", "Timestamp", "path"]

    ikeep = [table[0].index(col) for col in keep]
    table = [[row[i] for i in ikeep] for row in table]

    if args.rpath:  # Rename path to rpath title
        table[0] = [c if c != "path" else "rpath" for c in table[0]]

    if not args.header:
        table = table[1:]

    if not table:
        print(f"No files under {args.path!r}. Check the path and the date")
        return

    table = head_tail_table(
        table,
        head=args.head,
        tail=args.tail,
        dots=True,
        header=args.header,
    )
    table = tabulate(table)
    print(table, flush=True)


def file_versions(config):
    args = config.cliconfig

    dstdb = DFBDST(config)
    versions = dstdb.file_versions(args.filepath, count_refs=args.ref_count)

    # Build output
    out = [f"file: {args.filepath!r}"]

    table = []
    if args.header:
        table.append(["Ref. Count", "Size", "ModTime", "Timestamp", "Real Path"])
    for item in versions:
        row = [str(item.get("ref_count", ""))]

        if args.human:
            size = "{:0.2f} {}".format(*human_readable_bytes(item["size"]))
        else:
            size = str(item["size"])

        if item["size"] < 0:
            size = f"D"
        if item.get("isref", False):
            size = f"{size} (R)"
        row.append(size)

        mtime = item.get("mtime", None)
        if not mtime:
            mtime = ""
        else:
            mtime = (
                timestamp_parser(mtime, aware=True)
                .astimezone()
                .strftime(f"{STRFTIME_FMT}")
            )
        row.append(mtime)

        ts = item["timestamp"]
        ts = timestamp_parser(ts)
        if args.timestamp_local:
            ts = ts.astimezone().strftime(f"{STRFTIME_FMT}%z")
        else:
            ts = ts.strftime(f"{STRFTIME_FMT}Z")
        row.append(ts)

        if args.real_path >= 2:
            row.append(rcpathjoin(config.dst, item["rpath"]))
        else:
            row.append(item["rpath"])

        table.append(row)

    if not args.ref_count:
        table = [row[1:] for row in table]
    if not args.real_path:
        table = [row[:-1] for row in table]

    if table:
        table = head_tail_table(
            table,
            head=args.head,
            tail=args.tail,
            dots=True,
            header=args.header,
        )
        out.append(tabulate(table))
    else:
        out.append("  **No such file**. Check the path")

    out = "\n".join(out)
    print(out, flush=True)


def _timestamps_query(config):
    args = config.cliconfig
    dstdb = DFBDST(config)

    db = dstdb.db()

    before = args.before
    after = args.after
    path = args.path

    cond_dict = {}
    cond_sqls = []

    if path:
        path = path.removesuffix("/").removeprefix("./")
        cond_sqls.append("items.apath LIKE :path")
        cond_dict = {"path": f"{path}/%"}

    if before:
        b0 = before
        before = timestamp_parser(
            before,
            aware=True,
            epoch=True,
            now=config.now.obj,
        )
        logger.debug(f"Interpreted before = {b0} as {before} (s)")
        cond_sqls.append("timestamp <= :before")
        cond_dict["before"] = before

    if after:
        a0 = after
        after = timestamp_parser(
            after,
            aware=True,
            epoch=True,
            now=config.now.obj,
        )
        logger.debug(f"Interpreted after = {a0} as {after} (s)")
        cond_sqls.append("timestamp >= :after")
        cond_dict["after"] = after

    cond_sql = "WHERE " + " AND ".join(cond_sqls) if cond_sqls else ""

    # See https://stackoverflow.com/a/31704068/3633154 for the CASE WHEN ...
    ts_query = db.execute(
        f"""
        SELECT 
            timestamp,
            COUNT(timestamp) AS num_total,
            SUM(CASE WHEN size < 0 THEN 1 ELSE 0 END) AS num_del,
            SUM(CASE WHEN isref = 1 THEN 1 ELSE 0 END) AS num_mv,
            SUM(CASE WHEN (size >= 0 AND (isref IS NULL OR isref = 0) ) 
                     THEN size 
                     ELSE 0 
                     END) AS size
        FROM items 
        {cond_sql}
        GROUP BY timestamp
        ORDER BY timestamp""",
        cond_dict,
    )
    return ts_query


def timestamps(config):
    args = config.cliconfig

    ts_query = _timestamps_query(config)
    table = []
    if args.header:
        table.append(["Timestamp", "Total", "Deleted", "Moved", "Size"])

    for item in ts_query:
        timestamp = item["timestamp"]
        ts = timestamp_parser(timestamp, aware=True)

        if args.timestamp_local:
            ts = ts.astimezone().strftime(f"{STRFTIME_FMT}%z")
        else:
            ts = ts.strftime(f"{STRFTIME_FMT}Z")

        row = [ts]
        row.extend(item[k] for k in ["num_total", "num_del", "num_mv"])
        if args.human:
            size = "{:0.2f} {}".format(*human_readable_bytes(item["size"]))
        else:
            size = str(item["size"])
        row.append(size)
        table.append(row)

    table = [[str(c) for c in row] for row in table]

    table = head_tail_table(
        table,
        head=args.head,
        tail=args.tail,
        dots=True,
        header=args.header,
    )

    print(tabulate(table), flush=True)


def summary(config):
    args = config.cliconfig

    before = args.before
    after = args.after
    path = args.path

    ts = list(_timestamps_query(config))

    res = {}

    res["Path"] = repr(path.removesuffix("/").removeprefix("./") if path else "/")
    res["After"] = repr(after) if after else "<<earliest>>"
    res["Before"] = repr(before) if before else "<<latest>>"
    res["Timestamps"] = str(len(ts))

    res["Total"] = str(sum(r["num_total"] for r in ts))
    res["Deleted"] = str(sum(r["num_del"] for r in ts))
    res["Moved"] = str(sum(r["num_mv"] for r in ts))

    res["Size"] = sum(r["size"] for r in ts)
    res["Size"] = (
        f"{res['Size']} ({human_readable_bytes(res['Size'],fmt=True,short=True)})"
    )

    print(tabulate([[f"{k}:", v] for k, v in res.items()]))


def timestamp_include_filters(config):
    args = config.cliconfig
    dstdb = DFBDST(config)
    db = dstdb.db()

    before = args.before
    after = args.after
    path = args.path

    conditions = []

    if path:
        path = path.removesuffix("/").removeprefix("./")
        conditions.append(("items.apath LIKE :path", {"path": f"{path}/%"}))

    if before:
        b0 = before
        before = timestamp_parser(
            before,
            aware=True,
            epoch=True,
            now=config.now.obj,
        )
        logger.debug(f"Interpreted before = {b0} as {before} (s)")
        conditions.append(("timestamp <= :before", {"before": before}))

    if after:
        a0 = after
        after = timestamp_parser(
            after,
            aware=True,
            epoch=True,
            now=config.now.obj,
        )
        logger.debug(f"Interpreted after = {a0} as {after} (s)")
        conditions.append(("timestamp >= :after", {"after": after}))

    cond = ""
    params = {}
    if conditions:
        cond = "WHERE " + " AND ".join(condition[0] for condition in conditions)
        params |= {k: v for condition in conditions for k, v in condition[1].items()}

    query = f"""
        SELECT DISTINCT timestamp
        FROM items
        {cond}
        ORDER BY timestamp
        """

    qres = db.execute(query, params)
    includes = []
    for item in qres:
        timestamp = item["timestamp"]
        ts = timestamp_parser(timestamp, aware=True)
        dt = time2all(ts).dt

        includes.extend(["--include", f"*.{dt}*"])  # may or may not have a dot after

    print(shlex.join(includes))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, sys, shutil
from pathlib import Path
import gzip as gz
import lzma as xz
import re
import subprocess
import json
import itertools
import shlex
from textwrap import dedent

p = os.path.abspath("../")
if p not in sys.path:
    sys.path.insert(0, p)

from dfb.cli import cli
from dfb.utils import smart_splitext
from dfb.backup import NoCommonHashError

# Local
import testutils

# testing
import pytest


@pytest.mark.parametrize("rename_method", ["reference", "copy"])
def test_main(rename_method):
    test = testutils.Tester(name="main")

    test.config["renames"] = "mtime"
    test.config["filter_flags"] = ["--filter", "- *.exc"]
    test.config["rename_method"] = rename_method

    # # Main Test -- Tracking

    test.write_config()

    # +
    test.write_pre("src/versions.txt", "versions 1.")
    test.write_pre(
        "src/unįçôde, spaces, symb°ls (!@#$%^&*) in €‹›ﬁﬂ‡°·—±",
        "did it work?",
    )
    test.write_pre("src/untouched.txt", "Do not modify")
    test.write_pre("src/mod_same-size.txt", "modify but not change size")
    test.write_pre("src/mod_diff_size.txt", "modify and change size")
    test.write_pre(
        "src/mod_same-size-mtime.txt",
        "modify but not change size OR modTime",
    )
    test.write_pre("src/touch.txt", "touch me", dt=-1000)
    test.write_pre("src/sub/move.txt", "move me")
    test.write_pre("src/skip.exc", "skip me")
    test.write_pre("src/delete.txt", "delete me")

    test.write_pre("src/two_mv-1.txt", "Two will move")
    shutil.copy2("src/two_mv-1.txt", "src/two_mv-2.txt")

    test.backup("-v", offset=1)
    # -

    diff = test.src_missing_in_dst(keys=("apath", "size"))
    assert diff == {frozenset({("apath", "skip.exc"), ("size", 7)})}  # skip

    # +
    # Modify
    test.write_pre("src/versions.txt", "versions 3..")
    test.write_pre("src/new_at_3.txt", "I was created at 3")

    test.write_post("src/mod_same-size.txt", "ModifY BuT NoT ChangE SizE")
    test.write_post("src/mod_diff_size.txt", "modify and change size --DONE")

    stat0 = os.stat("src/mod_same-size-mtime.txt")
    test.write_post(
        "src/mod_same-size-mtime.txt", "MODIFY but Not change size OR modTime"
    )
    os.utime("src/mod_same-size-mtime.txt", (stat0.st_atime, stat0.st_mtime))

    Path("src/touch.txt").touch()
    test.move("src/sub/move.txt", "src/sub2/moved.txt")
    os.unlink("src/delete.txt")

    test.move("src/two_mv-1.txt", "src/two_moved-1.txt")
    test.move("src/two_mv-2.txt", "src/two_moved-2.txt")

    back = test.backup(offset=3)
    log = test.logs[-1][0]

    # -

    diff = test.src_missing_in_dst(keys="apath")

    assert diff == {
        frozenset({("apath", "mod_same-size-mtime.txt")}),
        frozenset({("apath", "skip.exc")}),
    }

    assert not any(
        ("apath", "new_at_3.txt") in f for f in test.remote_snapshot(before=2)
    )
    assert any(("apath", "new_at_3.txt") in f for f in test.remote_snapshot(before=4))

    # Test the move
    refitem = test.dstdb.snapshot(
        conditions=[("apath = :pp", {"pp": "sub2/moved.txt"})]
    ).fetchone()
    refitem = test.dstdb.fullrow2dict(refitem)

    if rename_method == "reference":
        r = json.loads(test.dst_rclone.read("sub2/moved.19700101000003R.txt"))
        assert r == {"ver": 2, "rel": "../sub/move.19700101000001.txt"}

        # DB
        assert refitem.get("isref", False)
        assert refitem["rpath"] == "sub/move.19700101000001.txt"
    else:
        assert test.read("dst/sub2/moved.19700101000003.txt") == "move me"

        # DB
        assert not refitem.get("isref", False)
        assert refitem.get("source_rpath", None) == "sub/move.19700101000001.txt"
        assert refitem.get("original", None) == "sub/move.txt"

    test.write_pre("src/versions.txt", "versions 5...")
    test.move("src/sub2/moved.txt", "src/moved_again.txt")
    test.backup("--refresh", offset=5)  # add refresh to test that too

    # Should still point to the original!
    if rename_method == "reference":
        r = json.loads(test.dst_rclone.read("moved_again.19700101000005R.txt"))
        assert r == {"ver": 2, "rel": "sub/move.19700101000001.txt"}
    else:
        test.read("dst/moved_again.19700101000005.txt") == "move me"

    diff = test.src_missing_in_dst(keys="apath")
    assert diff == {
        frozenset({("apath", "mod_same-size-mtime.txt")}),
        frozenset({("apath", "skip.exc")}),
    }

    assert "Too many matches for 'two_moved-1.txt'. Not moving" in log
    assert "Too many matches for 'two_moved-2.txt'. Not moving" in log

    # # Secondary Tests

    # ## Does a refresh look like the original

    # + tags=[]
    test.call("snapshot", "--output", "A.jsonl")

    # Use this opportunity to also test disable_refresh. -vv makes it throw error
    try:
        test.call("refresh", "-o", "disable_refresh = True", "-vv")
        assert False
    except ValueError:
        pass

    test.call("refresh")
    test.call("snapshot", "--output", "B.jsonl")

    with open("A.jsonl") as fp:
        A = [json.loads(line) for line in fp]
    with open("B.jsonl") as fp:
        B = [json.loads(line) for line in fp]

    # no mtime since precision issues
    keep = {"rpath", "apath", "timestamp", "size", "ref_rpath"}
    A = [{k: v for k, v in item.items() if k in keep} for item in A]
    B = [{k: v for k, v in item.items() if k in keep} for item in B]
    A = {frozenset(item.items()) for item in A}
    B = {frozenset(item.items()) for item in B}
    assert A == B

    ## Test reference format v2
    test.call("refresh", "-vv", "--no-use-snapshot")
    log = test.logs[-1][0]
    if rename_method == "reference":
        assert "Reference 'moved_again.19700101000005R.txt' is v2" in log

        # test format v1
        with open("dst/moved_again.19700101000005R.txt", "wt") as fp:
            fp.write("sub/move.19700101000001.txt")
        test.call("refresh", "-vv", "--no-use-snapshot")
        log = test.logs[-1][0]
        assert "Reference 'moved_again.19700101000005R.txt' is v1 (implied)" in log

        # Do it again with snapshot
        test.call("refresh", "-vv")
        log = test.logs[-1][0]
        assert (
            "Updated reference for 'moved_again.19700101000005R.txt' "
            "from 'sub/move.19700101000001.txt"
        ) in log

    # ## Do the snapshots match the reality?

    # Now that we are done, check all of the snapshots as read using the different tools
    # This tests the historic snapshots agains what they were.
    for ts, loc in test.backup_local_files.items():
        rem = test.remote_snapshot(before=ts + 0.1)
        miss_rem = {
            dict(a)["apath"] for a in loc - rem
        }  # Just the paths, but the whole thing has to agree
        miss_loc = {dict(a)["apath"] for a in rem - loc}  # ...

        assert miss_rem - {"skip.exc", "mod_same-size-mtime.txt"} == set()
        assert miss_loc - {"mod_same-size-mtime.txt"} == set()

    # ## Does restore match reality?
    #
    # Do a bona-fide restore and see

    for ts, loc in test.backup_local_files.items():
        test.call("restore", "--at", f"u{ts+0.1}", str(test.pwd / f"ts{ts}"))
        rem = test.local_files(test.pwd / f"ts{ts}")

        miss_rem = {
            dict(a)["apath"] for a in loc - rem
        }  # Just the paths, but the whole thing has to agree
        miss_loc = {dict(a)["apath"] for a in rem - loc}  # ...

        assert miss_rem - {"skip.exc", "mod_same-size-mtime.txt"} == set()
        assert miss_loc - {"mod_same-size-mtime.txt"} == set()

    # ## Do restore scripts match?
    #
    # Same as above but using the restore script

    for ts, loc in test.backup_local_files.items():
        restore_dir = str(test.pwd / f"script_ts{ts}")
        restore_script = str(test.pwd / f"script_ts{ts}.sh")
        test.call(
            "restore",
            "--at",
            f"u{ts+0.1}",
            restore_dir,
            "--shell-script",
            restore_script,
        )
        subprocess.check_call(["bash", restore_script])

        rem = test.local_files(restore_dir)

        miss_rem = {
            dict(a)["apath"] for a in loc - rem
        }  # Just the paths, but the whole thing has to agree
        miss_loc = {dict(a)["apath"] for a in rem - loc}  # ...

        assert miss_rem - {"skip.exc", "mod_same-size-mtime.txt"} == set()
        assert miss_loc - {"mod_same-size-mtime.txt"} == set()

    # ## Misc
    #
    # just to test some code paths
    test.call("restore-file", "sub/move.txt", ".", "--dry-run", "--at", "u1")
    test.call("restore-file", "sub/move.txt", ".", "--at", "1970-01-01 00:00:01Z")
    test.call(
        "restore-file",
        "sub/move.txt",
        "-",
        "--at",
        "1970-01-01 00:00:01Z",
        "--shell-script",
        "-",
    )

    out = test.ls("-d", "--after", "u4", "--no-header")
    items = {l.strip() for l in out.split("\n") if l.strip()}
    assert items == {"versions.txt", "moved_again.txt", "sub2/"}

    cli(["init", "new.py"])
    cli(["init", "new.py"])
    cli(["init", "new.py", "--force-overwrite"])

    test.config["rclone_env"]["RCLONE_CONFIG_PASS"] = "secret"
    test.write_config()
    r = repr(test.config_obj)
    assert "secret" not in r
    assert "**REDACTED**" in r


def test_shell():
    test = testutils.Tester(name="shell")

    # Configure each of the shell types. Need two tries...

    # +
    test.config["pre_shell"] = dedent(
        """\
            echo PRESHELL
            echo PWD = $PWD
            echo CONFIGDIR = $CONFIGDIR"""
    )
    test.config["post_shell"] = [
        "python",
        "-c",
        dedent(
            """\
        print("Post Shell")
        import os,sys
        print(f"{os.getcwd() = }")
        sys.exit(10)"""
        ),
    ]

    test.config["stop_on_shell_error"] = False
    # -

    test.write_config()

    test.write_pre("src/file.txt", "file")
    test.backup(offset=1)

    # +
    out = test.logs[-1][0]
    assert "pre.shell: $ echo PRESHELL" in out
    assert "pre.shell: $ echo PWD = $PWD" in out
    assert "pre.shell: $ echo CONFIGDIR = $CONFIGDIR" in out
    assert "out: PRESHELL" in out
    assert "PWD = /" in out
    assert "CONFIGDIR = /" in out

    assert (
        r"""post.shell ['python', '-c', 'print("Post Shell")\nimport os,sys\nprint(f"{os.getcwd() = }")\nsys.exit(10)']"""
        in out
    )

    assert "out: Post Shell" in out
    assert "os.getcwd() = '/" in out
    # -

    test.config["post_shell"] = dict(
        cmd=[
            "python",
            "-c",
            dedent(
                """\
                import os
                print(f"{os.getcwd() = }")
                print(f"{os.environ.get('SHELL_TEST','FAIL') = }")
                print(f"{os.environ.get('CONFIGDIR','FAIL') = }")
                print(f"{os.environ.get('STATS','FAIL') = }")
                """
            ),
        ],
        shell=False,
        cwd=os.path.expanduser("~/"),
        env={"SHELL_TEST": "SUCCESS"},
    )
    test.config["pre_shell"] = ""
    test.write_config()

    test.write_pre("src/file.txt", "file3")
    test.backup(offset=3)

    out = test.logs[-1][0]
    assert "os.getcwd() = '/" in out
    assert "os.environ.get('SHELL_TEST','FAIL') = 'SUCCESS'" in out
    assert "os.environ.get('CONFIGDIR','FAIL') = '/" in out


def test_log_upload():
    test = testutils.Tester(name="log_upload")

    test.write_config()

    test.write_pre("src/versions.txt", "versions 1.")
    test.backup(offset=1)
    assert os.path.exists("dst/.dfb/logs/19700101000001Z.log")

    test.write_pre("src/versions.txt", "versions 2..")
    test.backup(offset=3, allow_error=True)
    assert os.path.exists("dst/.dfb/logs/19700101000003Z.log")


def test_dst_compare_and_dst_renames():
    # Test and verify using local vs destination attributes
    test = testutils.Tester(name="dst_attributes")

    # ## Compare
    #
    # See how comparisons go with and without reusing hashes

    vq = ["-v"]

    # +
    test.config["dst_list_rclone_flags"] = ["--fast-list"]

    test.config["compare"] = "hash"
    test.config["dst_compare"] = "hash"
    test.write_config()
    # -

    test.write_pre("src/file.txt", "file", dt=-30)
    back = test.backup(*vq, offset=1)

    test.write_pre("src/file.txt", "file", dt=-28)
    back = test.backup(*vq, offset=3)
    assert back.new + back.modified == []  # No transfer

    test.write_pre("src/file.txt", "file", dt=-26)
    test.backup("--refresh", *vq, offset=5)
    assert back.new + back.modified == []  # No transfer

    test.config["dst_compare"] = "mtime"
    test.write_config()

    # This should still not transfer since we are src-to-src compare
    test.write_pre("src/file.txt", "file", dt=-24)
    back = test.backup(*vq, offset=7)
    assert back.new + back.modified == []  # No transfer

    # This should transfer since we are src-to-dst compare
    test.write_pre("src/file.txt", "file", dt=-22)
    back = test.backup("--refresh", "--no-refresh-use-snapshots", "-v", offset=9)
    assert back.new + back.modified == ["file.txt"]  # Transfer

    # Side Test: Make sure dst_list_rclone_flags got set. Look for --fast-list
    log = test.logs[-1][0]
    assert re.search("rclone call.*--fast-list", log)

    # This again NOT transfer since we are back to src-to-src
    test.write_pre("src/file.txt", "file", dt=-20)
    back = test.backup(*vq, offset=11)
    assert back.new + back.modified == []  # No transfer

    # ## Moves

    # +
    test.config["compare"] = "mtime"
    test.config["dst_compare"] = "mtime"

    test.config["renames"] = "hash"
    test.config["dst_renames"] = "hash"
    test.write_config()
    # -

    test.write_pre("src/mv1.txt", "file", dt=-30)
    back = test.backup(*vq, offset=13)

    test.move("src/mv1.txt", "src/mv2.txt")
    back = test.backup(*vq, offset=15)
    assert len(back.moves) == 1

    test.move("src/mv2.txt", "src/mv3.txt")
    back = test.backup("--refresh", *vq, offset=17)
    assert len(back.moves) == 1

    test.config["dst_renames"] = "mtime"
    test.write_config()

    # +
    # This should still track the move because we are src-to-src
    test.move("src/mv3.txt", "src/mv4.txt")
    back = test.backup(*vq, offset=19)

    assert len(back.moves) == 1
    assert (
        "Compare 'mv4.txt' with attrib = 'hash'. MATCH"
        in test.logs[-1][0] + test.logs[-1][1]
    )

    # +
    # This should still track the move from mtime because we are src-to-dst
    test.move("src/mv4.txt", "src/mv5.txt")
    back = test.backup("--refresh", "--no-refresh-use-snapshots", *vq, offset=21)

    assert len(back.moves) == 1
    assert (
        "Compare 'mv5.txt' with attrib = 'mtime'. MATCH"
        in test.logs[-1][0] + test.logs[-1][1]
    )
    # -

    test.config["dst_renames"] = False
    test.write_config()

    test.move("src/mv5.txt", "src/mv6.txt")
    back = test.backup(*vq, offset=23)
    assert len(back.moves) == 1
    assert (
        "Compare 'mv6.txt' with attrib = 'hash'. MATCH"
        in test.logs[-1][0] + test.logs[-1][1]
    )

    test.move("src/mv6.txt", "src/mv7.txt")
    back = test.backup("--refresh", "--no-refresh-use-snapshots", *vq, offset=25)
    # Disabled. Check it
    assert len(back.moves) == 0
    assert back.new == ["mv7.txt"]


def test_restore_error():
    test = testutils.Tester(name="restore_error")

    test.write_config()

    test.write_pre("src/file1.txt", "file1")
    test.write_pre("src/sub/file2.txt", "file2")
    test.backup("-q", offset=1)

    test.write_pre("src/file3.txt", "file3")
    test.write_pre("src/sub/file2.txt", "file2.")
    test.backup("-q", offset=3)

    # +
    test.call("restore-file", "sub/file2.txt", "testfile.txt", "--to", "-q")
    assert test.read("testfile.txt") == "file2."

    test.call(
        "restore-file",
        "sub/file2.txt",
        "testfile.txt",
        "--to",
        "--no-check",
        "--at",
        "u2",
        "-q",
    )
    assert test.read("testfile.txt") == "file2"
    # -

    # While we are at it, restore to the source to test that
    test.call("restore-dir", "@src/new", "--no-check")

    assert {dict(i)["apath"] for i in test.local_files()} == {
        "file1.txt",
        "file3.txt",
        "new/file1.txt",
        "new/file3.txt",
        "new/sub/file2.txt",
        "sub/file2.txt",
    }

    test.backup(offset=5)

    os.unlink("dst/sub/file2.19700101000001.txt")

    test.call(
        "restore-file", "sub/file2.txt", "testfile.txt", "--to", "-q", "--no-check"
    )
    assert test.read("testfile.txt") == "file2."

    # +
    test.call(
        "restore-file",
        "sub/file2.txt",
        "testfile.txt",
        "--to",
        "--no-check",
        "--at",
        "u2",
    )  # must not have -q

    log = test.logs[-1][0]
    assert "ERROR: Could not restore 'sub/file2.19700101000001.txt'." in log
    assert "At least one restore did not work" in log
    assert test.read("testfile.txt") != "file2"  # What is should be
    assert test.read("testfile.txt") == "file2."  # from before
    # -

    test.call("restore-dir", "@src/neww", "--no-check", "--at", "u2")
    assert "ERROR: Could not restore 'sub/file2.19700101000001.txt'." in log
    assert "At least one restore did not work." in log

    test.call("versions", "sub/file2.txt")
    test.call("refresh")
    test.call(
        "restore-file",
        "sub/file2.txt",
        "testfile.txt",
        "--to",
        "--no-check",
        "--at",
        "u2",
    )  # must not have -q
    log = test.logs[-1][0]
    assert "ERROR: Could not find 'sub/file2.txt' at the specified time" in log

    # Test some restores with no file


@pytest.mark.parametrize("mode", ["size", "mtime", "hash"])
def test_false_negs_compare(mode):
    test = testutils.Tester(name="false_negs_compare")

    compare = mode

    test.config["compare"] = compare
    test.write_config()

    # +
    test.write_pre("src/no_size_change.txt", "123")

    test.write_pre("src/no_size-mtime_change.txt", "1234")
    t0 = os.stat("src/no_size-mtime_change.txt")

    test.write_pre("src/touch.txt", "touch me")

    test.write_pre("src/change_size.txt", "ABCD")
    t1 = os.stat("src/change_size.txt")

    test.backup(offset=1)

    # +
    test.write_post("src/no_size_change.txt", "321")

    test.write_post("src/no_size-mtime_change.txt", "4321")
    os.utime("src/no_size-mtime_change.txt", (t0.st_atime, t0.st_mtime))

    test.write_post("src/touch.txt", "touch me")

    test.write_post("src/change_size.txt", "ABC")
    os.utime("src/change_size.txt", (t1.st_atime, t1.st_mtime))

    back = test.backup("-v", offset=3)

    if compare == "size":
        assert set(back.modified) == {"change_size.txt"}
    elif compare == "mtime":
        assert set(back.modified) == {
            "change_size.txt",
            "no_size_change.txt",
            "touch.txt",
        }
    elif compare == "hash":
        assert set(back.modified) == {
            "change_size.txt",
            "no_size_change.txt",
            "no_size-mtime_change.txt",
        }


def test_missing_ref():
    test = testutils.Tester(name="missing_ref")

    test.write_config()

    # +
    test.write_pre("src/file1.txt", "123")
    test.write_pre("src/file2.txt", "ABCD")

    test.backup(offset=1)

    # +
    test.move("src/file1.txt", "src/fileONE.txt")
    test.write_pre("src/file2.txt", "ABCDE")

    test.backup(offset=3)
    # -

    os.unlink("dst/file1.19700101000001.txt")

    test.call("ls")

    test.call("restore-file", "file2.txt", "-", "-q")

    # +
    test.call("restore-file", "fileONE.txt", "-")
    log = test.logs[-1][0]
    assert "ERROR: Could not restore 'file1.19700101000001.txt'" in log
    assert "At least one restore did not work" in log

    # Same thing but catch the error
    try:
        test.call("restore-file", "fileONE.txt", "-", "-v")  # -v will make it error
        assert False
    except ValueError:
        pass
    # -

    test.call("refresh", "--no-use-snapshot")
    out = test.ls("-q")

    log = "\n".join(l[0] for l in test.logs[-2:])
    assert (
        "WARNING: File 'fileONE.19700101000003R.txt' "
        "references 'file1.19700101000001.txt' but it is missing. "
        "Will just be treated as deleted"
    ) in log

    assert "file2.txt" in out

    out = test.ls("-d")
    assert "fileONE.txt (DEL)" in out


def test_override():
    """
    Test overrides including pre,post values.

    Do this simply by adding some print statements
    """
    test = testutils.Tester(name="override")
    test.write_config()

    test.write_pre("src/file1.txt", "1")
    test.backup("-o", "print('twice?')", offset=1)
    log = test.logs[-1][0]
    assert log.count("twice?") == 2

    test.write_pre("src/file1.txt", "12")
    test.backup(
        "-o",
        dedent(
            """\
        if pre:
            print('one PRE')
            try:
                print(f"{newval = }")
                raise ValueError()
            except NameError:
                newval = 10
                print('Set newval')
        if post:
            print('one POST')
            print(f"{newval = }")
        """
        ),
        offset=3,
    )
    log = test.logs[-1][0]
    assert log.count("one PRE") == 1
    assert log.count("one POST") == 1
    assert log.count("Set newval") == 1
    assert log.count("newval = 10") == 1


def test_subdirs():
    test = testutils.Tester(name="subdirs")

    filters = [
        "- *.exc",  # Wildcard
        "- general_exc.txt",  # General
        "- /sub/specific_exc.txt",  # specific. Anchored
        "- /sub2/**",  # General
    ]

    test.config["filter_flags"] = []
    for filt in filters:
        test.config["filter_flags"].extend(["--filter", filt])

    test.write_config()

    test.write_pre("src/file1.txt", "1")
    test.write_pre("src/sub/file2.txt", "2")
    test.write_pre("src/sub/file3.exc", "3")
    test.write_pre("src/sub/general_exc.txt", "exc")
    test.write_pre("src/sub/specific_exc.txt", "exc")
    test.write_pre("src/file4.exc", "4")
    test.write_pre("src/sub2/file5.txt", "5")

    test.backup(offset=1)

    assert test.src_missing_in_dst(keys=("apath",)) == {
        frozenset({("apath", "sub/file3.exc")}),
        frozenset({("apath", "sub/specific_exc.txt")}),
        frozenset({("apath", "sub/general_exc.txt")}),
        frozenset({("apath", "file4.exc")}),
        frozenset({("apath", "sub2/file5.txt")}),
    }

    test.write_post("src/file1.txt", "1POST")
    test.write_post("src/sub/file2.txt", "2POST")

    back = test.backup("--subdir", "sub", offset=3)

    assert set(back.new) == {"sub/specific_exc.txt"}  # Broken specific filter
    assert set(back.modified) == {"sub/file2.txt"}  # No file1.txt

    back = test.backup(offset=5)

    assert set(back.modified) == {"file1.txt"}  # Now captures the mod
    assert set(back.deleted) == {
        "sub/specific_exc.txt"
    }  # removed since it is now filtered

    # Make sure we never made sub2
    assert not os.path.exists("dst/sub2")


def test_subdir_w_empty():
    """
    Regression for non-empty (or maybe even actually empty) subdirs w/o modified
    files showing as empty
    """
    test = testutils.Tester(name="subdir_w_empty")

    test.config["empty_directory_markers"] = True
    test.write_config()

    test.write_pre("src/file1.txt", "1")
    test.write_pre("src/sub0/file2.txt", "2")
    test.write_pre("src/sub0/sub1/file3.txt", "3")
    test.write_pre("src/sub0/sub2/file4.txt", "4")

    test.backup(offset=1)

    test.write_post("src/sub0/sub2/file4.txt", "four")
    os.makedirs("src/sub0/sub3")

    test.backup("--subdir", "sub0", offset=3)

    # Make sure the empty dir maker is created even at the subdir
    assert os.path.exists("dst/sub0/sub3/.dfbempty.19700101000003")

    # this is the regression. Notice these are relative to 'sub0'
    assert not os.path.exists("dst/" + "sub1/.dfbempty.19700101000003")
    assert not os.path.exists("dst/" + "sub2/.dfbempty.19700101000003")


@pytest.mark.parametrize("mode", ["link", "link-webdav", "copy", "skip"])
def test_symlinks(mode):
    test = testutils.Tester(name="symlinks", src="srcalias:")

    linkmode = mode
    args = []
    if mode == "link-webdav":
        linkmode = "link"
        webdav = testutils.WebDAV(path="dst").start()
        args = ["--override", f"dst = {webdav.remote!r}"]

    # test.config["links"] = linkmode
    if linkmode == "link":
        test.config["rclone_flags"] = ["--links"]
    elif linkmode == "copy":
        test.config["rclone_flags"] = ["--copy-links"]
    elif linkmode == "skip":
        test.config["rclone_flags"] = ["--skip-links"]

    test.config["upload_logs"] = False

    test.write_config()

    test.write("src/file1.txt", "File ONE")
    test.write("src/sub/file2.txt", "File 2")
    os.symlink("file1.txt", "src/link1.txt")
    os.symlink("sub/file2.txt", "src/link2.1.txt")
    os.symlink("file2.txt", "src/sub/link2.2.txt")
    Path("src/other/").mkdir(exist_ok=True, parents=True)
    os.symlink("../file1.txt", "src/other/backfile1.txt")
    os.symlink("../sub/file2.txt", "src/other/backfile2.txt")

    test.backup(*args, offset=1)
    test.call("restore", "res", *args)
    test.call("refresh")

    if mode == "link":
        assert os.readlink("dst/link1.19700101000001.txt") == "file1.txt"
        assert os.readlink("dst/link2.1.19700101000001.txt") == "sub/file2.txt"
        assert os.readlink("dst/other/backfile1.19700101000001.txt") == "../file1.txt"
        assert (
            os.readlink("dst/other/backfile2.19700101000001.txt") == "../sub/file2.txt"
        )
        assert os.readlink("dst/sub/link2.2.19700101000001.txt") == "file2.txt"

        # Restore test. Do not need to do them all. Just make sure it's a link
        assert os.readlink("res/link1.txt") == "file1.txt"

    elif mode == "link-webdav":
        assert {dict(a)["apath"] for a in test.remote_snapshot()} == {
            "file1.txt",
            "link1.txt.rclonelink",
            "link2.1.txt.rclonelink",
            "other/backfile1.txt.rclonelink",
            "other/backfile2.txt.rclonelink",
            "sub/file2.txt",
            "sub/link2.2.txt.rclonelink",
        }
        assert set(testutils.tree("dst")) == {
            "dst/file1.19700101000001.txt",
            "dst/link1.19700101000001.txt.rclonelink",
            "dst/link2.1.19700101000001.txt.rclonelink",
            "dst/other/backfile1.19700101000001.txt.rclonelink",
            "dst/other/backfile2.19700101000001.txt.rclonelink",
            "dst/sub/file2.19700101000001.txt",
            "dst/sub/link2.2.19700101000001.txt.rclonelink",
        }

        assert test.read("dst/link1.19700101000001.txt.rclonelink") == "file1.txt"
        assert test.read("dst/link2.1.19700101000001.txt.rclonelink") == "sub/file2.txt"
        assert (
            test.read("dst/other/backfile1.19700101000001.txt.rclonelink")
            == "../file1.txt"
        )
        assert (
            test.read("dst/other/backfile2.19700101000001.txt.rclonelink")
            == "../sub/file2.txt"
        )
        assert test.read("dst/sub/link2.2.19700101000001.txt.rclonelink") == "file2.txt"
        assert len(test.tree_sha1s("dst")) == 7

        # Restore test. Do not need to do them all. Just make sure it's a link
        assert os.readlink("res/link1.txt") == "file1.txt"

    elif mode == "copy":
        assert {dict(a)["apath"] for a in test.remote_snapshot()} == {
            "file1.txt",
            "link1.txt",
            "link2.1.txt",
            "other/backfile1.txt",
            "other/backfile2.txt",
            "sub/file2.txt",
            "sub/link2.2.txt",
        }
        assert set(testutils.tree("dst")) == {
            "dst/file1.19700101000001.txt",
            "dst/link1.19700101000001.txt",
            "dst/link2.1.19700101000001.txt",
            "dst/other/backfile1.19700101000001.txt",
            "dst/other/backfile2.19700101000001.txt",
            "dst/sub/file2.19700101000001.txt",
            "dst/sub/link2.2.19700101000001.txt",
        }
        assert not any(os.path.islink(f) for f in testutils.tree("dst"))
        assert len(test.tree_sha1s("dst")) == 2

        assert not os.path.islink("res/link1.txt")

    elif mode == "skip":
        assert {dict(a)["apath"] for a in test.remote_snapshot()} == {
            "file1.txt",
            "sub/file2.txt",
        }
        assert set(testutils.tree("dst")) == {
            "dst/file1.19700101000001.txt",
            "dst/sub/file2.19700101000001.txt",
        }
        assert len(test.tree_sha1s("dst")) == 2

        assert not os.path.exists("res/link1.txt")

    # Make sure a second copy doesn't mess with this
    back2 = test.backup(offset=3)
    assert back2.modified == back2.moves == back2.deleted == []

    back2 = test.backup("--refresh", offset=5)
    assert back2.modified == back2.moves == back2.deleted == []


def test_symlinks_in_union():
    """
    This is a (regression) test for links in a union drive
    """
    test = testutils.Tester(name="symlink_in_union")
    test.config["links"] = "link"
    test.config["concurrency"] = 1
    test.config["upload_logs"] = False

    os.makedirs("empty")
    empty = os.path.abspath("empty")
    src = test.config["src"]

    env = test.config["rclone_env"]
    env["RCLONE_UNION_UPSTREAMS"] = shlex.join([src, empty])

    test.config["src"] = ":union:"
    test.write_config()

    test.write("src/file1.txt", "File ONE")
    os.symlink("file1.txt", "src/link1.txt")
    test.backup("-v", offset=1)


def test_snapshots():
    test = testutils.Tester(name="snapshots")

    test.write_config()

    test.write_pre("src/versions.txt", "versions 1")
    test.write_pre("src/new1.txt", "do not touch")
    test.write_pre("src/new1_del3.txt", "new at 1. Del at 3")

    test.backup(offset=1)

    test.write_post("src/versions.txt", "versions 3.")
    os.unlink("src/new1_del3.txt")
    test.write_pre("src/new3_mv5.txt", "new at 3. mv at 5")

    test.backup(offset=3)

    test.write_pre("src/versions.txt", "versions 5..")
    test.move("src/new3_mv5.txt", "src/MOVED/new3_mv5 DONE.txt")

    test.backup(offset=5)

    # Call for snapshots. Exports should be cumulative
    test.call("snapshot", "--output", "1.jsonl", "--only", "u1", "--deleted")
    test.call("snapshot", "--output", "3.jsonl", "--only", "u3", "--deleted")
    test.call("snapshot", "--output", "5.jsonl", "--only", "u5", "--deleted")
    test.call("snapshot", "--output", "e1.jsonl", "--before", "u1", "--export")
    test.call("snapshot", "--output", "e3.jsonl", "--before", "u3", "--export")
    test.call("snapshot", "--output", "e5.jsonl", "--before", "u5", "--export")

    keys = ["rpath", "apath", "timestamp", "size"]
    cumupl = set()
    for uz in [1, 3, 5]:
        with open(f"{uz}.jsonl") as fp:
            cli = [json.loads(line) for line in fp]
            cli = {frozenset((k, f[k]) for k in keys) for f in cli}

        with open(f"e{uz}.jsonl") as fp:
            exp = [json.loads(line) for line in fp]
            exp = {frozenset((k, f[k]) for k in keys) for f in exp}

        with gz.open(f"dst/.dfb/snapshots/1970/01/1970010100000{uz}Z.jsonl.gz") as fp:
            upl = [json.loads(line) for line in fp]
            upl = {frozenset((k, f[k]) for k in keys) for f in upl}
            cumupl.update(upl)

        assert cli == upl
        assert exp == cumupl

    # Test with import
    test.call("snapshot", "--output", "oe1.jsonl", "--only", "u1", "--export")
    test.call("snapshot", "--output", "oe3.jsonl", "--only", "u3", "--export")
    test.call("snapshot", "--output", "oe5.jsonl", "--only", "u5", "--export")

    # w/ reset
    test.call("advanced", "dbimport", "--reset", "oe1.jsonl", "oe3.jsonl", "oe5.jsonl")
    test.call("snapshot", "--output", "new.jsonl", "--export")
    with open("new.jsonl") as fp:
        exp = [json.loads(line) for line in fp]
        exp = {frozenset((k, f[k]) for k in keys) for f in exp}
    assert exp == cumupl

    # w/ reset then each its own all. Should be the same
    test.call("advanced", "dbimport", "--reset")  # reset
    test.call("advanced", "dbimport", "oe1.jsonl")
    test.call("advanced", "dbimport", "oe3.jsonl")
    test.call("advanced", "dbimport", "oe5.jsonl")
    test.call("snapshot", "--output", "new.jsonl", "--export")
    with open("new.jsonl") as fp:
        exp = [json.loads(line) for line in fp]
        exp = {frozenset((k, f[k]) for k in keys) for f in exp}
    assert exp == cumupl

    # w/ reset but in a directory
    Path("tmpdir").mkdir()
    shutil.copy2("oe1.jsonl", "tmpdir")
    shutil.copy2("oe3.jsonl", "tmpdir")
    shutil.copy2("oe5.jsonl", "tmpdir")
    test.call("advanced", "dbimport", "--reset", "--dirs", "tmpdir")
    test.call("snapshot", "--output", "new.jsonl", "--export")
    with open("new.jsonl") as fp:
        exp = [json.loads(line) for line in fp]
        exp = {frozenset((k, f[k]) for k in keys) for f in exp}
    assert exp == cumupl

    # w/o reset
    test.call("advanced", "dbimport", "oe1.jsonl", "oe3.jsonl", "oe5.jsonl")
    test.call("snapshot", "--output", "new.jsonl", "--export")
    with open("new.jsonl") as fp:
        exp = [json.loads(line) for line in fp]
        exp = {frozenset((k, f[k]) for k in keys) for f in exp}
    assert exp == cumupl

    ## Now prune things
    test.call("prune", "u7", offset=7)
    shutil.copy2(
        "dst/.dfb/snapshots/1970/01/19700101000007Z.jsonl.gz", "prune.jsonl.gz"
    )

    with gz.open("prune.jsonl.gz", "rt") as fp:
        for line in fp:
            line = json.loads(line)
            assert line["_V"] == 1
            assert line["_action"] == "prune"

    # reset, import again then apply.
    # Also test the uploads
    test.call(
        "advanced",
        "dbimport",
        "--reset",
        "oe1.jsonl",
        "oe3.jsonl",
        "oe5.jsonl",
        "prune.jsonl.gz",
        "--upload",
        offset=1_234_567_890,
    )
    log = test.logs[-1][0]
    assert "Imported 0 files and will prune 4" in log
    assert "Imported 3 files and will prune 4" not in log, "should not have mixed"
    assert "Pruned 4 files from all exports" in log

    test.call("snapshot", "--output", "new.jsonl", "--export")
    with open("new.jsonl") as fp:
        exp = [json.loads(line) for line in fp]
        exp = {frozenset((k, f[k]) for k in keys) for f in exp}
    assert exp != cumupl, "Should NOT match with the exports because it was pruned!"

    # test that the files got uploaded. Should NOT be compressed unless already
    for ff in ["0.oe1.jsonl", "1.oe3.jsonl", "2.oe5.jsonl", "3.prune.jsonl.gz"]:
        assert (test.pwd / "dst/.dfb/snapshots/2009/02/20090213233130Z" / ff).exists()


def test_missing_hashes():
    test = testutils.Tester(name="missing_hashes")
    test.config["compare"] = "hash"
    test.write_config()

    test.write_pre("src/same_size.txt", "versions 1")
    test.backup(offset=1)

    try:
        from dfb import _FAIL

        _FAIL.add("missing_hashes")

        test.write_post("src/same_size.txt", "versions 2")  # Same size!
        test.backup(offset=3)
        log = test.logs[-1][0]
        assert "WARNING: Missing hashes on source and/or dest" in log
        assert "Reverting to 'size' only" in log

        # Make sure it didn't back it up
        assert not os.path.exists("dst/same_size.19700101000003.txt")

        # This one should work even with missing hashes
        test.backup("-o", "compare='mtime'", offset=5)
        assert os.path.exists("dst/same_size.19700101000005.txt")

        # Do it again but this should fail
        # '-v' to get an error
        try:
            test.backup("-v", "-o", "error_on_missing_hash = True", offset=7)
            assert False
        except NoCommonHashError:
            pass

    finally:
        _FAIL.remove("missing_hashes")


@pytest.mark.parametrize("metadata", [True, False])
def test_metadata(metadata):
    test = testutils.Tester(name="metadata")
    test.config["metadata"] = metadata

    # Make sure macOS isn't cloning as that won't test the metadata.
    test.config["rclone_flags"] = test.config.get("rclone_flags", [])
    test.config["rclone_flags"].append("--local-no-clone")
    test.write_config()

    uu = 4, 6, 7
    gg = oo = tuple(range(8))
    for u, g, o in itertools.product(uu, gg, oo):
        filename = f"{u}{g}{o}.txt"
        perm = int(f"{u}{g}{o}", 8)
        test.write_pre(f"src/{filename}", filename)
        os.chmod(f"src/{filename}", perm)

    test.backup(offset=1)
    test.call("restore", "--at", "u1.1", str(test.pwd / f"res1"))

    # This is slow so I comment it out but it has been tested and I can come back to it.
    # restore_dir = str(test.pwd / "res2")
    # restore_script = str(test.pwd / "res2.sh")
    # test.call("restore","--at","u1.1",restore_dir,"--shell-script",restore_script)
    # subprocess.check_call(["bash", restore_script])

    # Verify the src too just to be complete
    n = 0
    correct = 0
    vdirs = ["src", "dst", "res1"]  # ,'res2']
    for vdir in vdirs:
        for file in os.listdir(vdir):
            gold = file.split(".", 1)[0]
            mode = os.stat(os.path.join(vdir, file)).st_mode
            correct += int(oct(mode).endswith(gold))
            n += 1
    print(f"verified {n}. Correct {correct}")

    if metadata:
        assert correct == n
    else:
        assert correct < n


def test_dump():
    test = testutils.Tester(name="dump")

    test.config["metadata"] = False  # Issues with atime due to run time
    test.config["renames"] = "mtime"
    test.write_config()

    def comp(t):
        """Compare results. Order doesn't matter"""
        with open(f"t{t}.jsonl") as fp:
            dump = {testutils.dict2frozen(json.loads(l)) for l in fp}
        with gz.open(f"dst/.dfb/snapshots/1970/01/197001010000{t:02d}Z.jsonl.gz") as fp:
            snap = {testutils.dict2frozen(json.loads(l)) for l in fp}
        return dump == snap

    test.write_pre("src/new.txt", "new!")
    test.write_pre("src/modify.txt", "modify")
    test.write_pre("src/move_by_ref.txt", "will move by ref")
    test.write_pre("src/move_by_ref_then_copy.txt", "will move by ref then copy")
    test.write_pre("src/move_by_copy.txt", "will move by copy")
    test.write_pre("src/delete.txt", "delete")

    test.backup("--dump", "t1.jsonl", offset=1)
    test.backup(offset=1)
    assert comp(1)

    test.write_post("src/modify.txt", "modify.")
    shutil.move("src/move_by_ref.txt", "src/moved_by_ref.txt")
    shutil.move("src/move_by_ref_then_copy.txt", "src/moved_by_ref_then_copy.txt")
    os.unlink("src/delete.txt")

    test.backup("--dump", "t3.jsonl.gz", offset=3)  # Just checking the compression
    test.backup("--dump", "t3.jsonl.xz", offset=3)  # Just checking the compression
    test.backup("--dump", "t3.jsonl", offset=3)
    test.backup(offset=3)
    assert comp(3)

    test.write_post("src/modify.txt", "modify..")
    shutil.move("src/moved_by_ref_then_copy.txt", "src/moved_by_ref_now_copy.txt")
    shutil.move("src/move_by_copy.txt", "src/moved_by_copy.txt")

    test.backup("--dump", "t5.jsonl", "-o", "rename_method = 'copy'", offset=5)
    test.backup("-o", "rename_method = 'copy'", offset=5)
    assert comp(5)

    with (
        open("t3.jsonl", "rt") as r,
        gz.open("t3.jsonl.gz", "rt") as g,
        xz.open("t3.jsonl.xz", "rt") as x,
    ):
        assert r.read() == g.read() == x.read()

    test.call("prune", "u4", "--dump", "t7.jsonl", offset=7)
    test.call("prune", "u4", offset=7)
    assert comp(7)


def test_auto():
    """
    WARNING -- these may have to be updated in newer versions of rclone if
               any of the remotes get new capabilities!
    """
    autos = [
        *["compare", "dst_compare"],
        *["renames", "dst_renames"],
        *["get_modtime", "get_hashes"],
    ]

    test = testutils.Tester(name="auto_settings")

    test.config |= {k: "auto" for k in autos}

    config = test.write_config()
    assert all(getattr(config, k, False) == "auto" for k in autos)

    config._set_auto()

    assert {k: getattr(config, k, None) for k in autos} == {
        "compare": "mtime",
        "dst_compare": "mtime",
        "renames": "mtime",
        "dst_renames": "mtime",
        "get_modtime": True,
        "get_hashes": False,
    }


def test_min_size():
    test = testutils.Tester(name="min_size")

    test.config["min_rename_size"] = "0.000005 mb"  # 5 bytes
    test.write_config()

    test.write_pre("src/small0.txt", "0")
    test.write_pre("src/large0.txt", "0123456789")

    test.backup(offset=1)

    shutil.move("src/small0.txt", "src/small1.txt")
    shutil.move("src/large0.txt", "src/large1.txt")

    test.backup("-v", offset=3)

    assert test.read("dst/small1.19700101000003.txt") == "0", "moved"
    assert not os.path.exists("dst/small1.19700101000003R.txt")

    assert json.loads(test.read("dst/large1.19700101000003R.txt")) == {
        "ver": 2,
        "rel": "large0.19700101000001.txt",
    }
    assert not os.path.exists("dst/large1.19700101000003.txt")

    # Make it call copy
    test.config["min_rename_size"] = "5 Pib"
    test.write_config()

    shutil.move("src/small1.txt", "src/small2.txt")
    shutil.move("src/large1.txt", "src/large2.txt")

    test.backup("-v", offset=5)

    assert test.read("dst/small2.19700101000005.txt") == "0"
    assert not os.path.exists("dst/small2.19700101000005R.txt")

    assert test.read("dst/large2.19700101000005.txt") == "0123456789"
    assert not os.path.exists("dst/large2.19700101000005R.txt")


def test_push_snapshots():
    """
    Test pushing snapshot files to the destination

    Note that testing for dbimport is done in tha
    """
    test = testutils.Tester(name="push_snap")
    test.write_config()

    test.write_pre("src/file.txt", "0")

    # Write a .jsonl file that mimics being left from last time. Should also be compressed
    tfile = test.pwd / "cache/DFB/test_push_snap.snap" / "2022/06" / "test.jsonl"
    tfile.parent.mkdir(exist_ok=True, parents=True)
    tfile.write_text("this is a dummy file")

    test.backup(offset=1)

    # Make sure it got backed up
    qfile = test.pwd / "dst/.dfb/snapshots/2022/06/test.jsonl.gz"

    assert not tfile.exists(), "didn't get moved"
    assert qfile.exists()
    assert not qfile.with_suffix("").exists(), "didn't get compressed"
    assert gz.open(str(qfile), "rt").read() == "this is a dummy file", "not compressed"


def test_empty_dirs():
    test = testutils.Tester(name="empty_dirs")

    test.config["filter_flags"] = ["--filter", "- *.exc"]
    test.config["empty_directory_markers"] = True

    test.write_config()

    os.makedirs("src/empty")
    os.makedirs("src/hassub/subdir/")

    test.backup("-v", offset=1)

    assert os.path.exists("dst/empty/.dfbempty.19700101000001")
    assert os.path.exists("dst/hassub/subdir/.dfbempty.19700101000001")

    shutil.move("src/hassub/subdir/", "src/hassub/subdir_moved")
    test.write_pre("src/empty/file.txt", "file")

    test.backup("-v", offset=3)

    # Test both now and past
    assert test.tree("--at", "u2") == dedent(
        """\
        /
        ├── empty/
        │   └── .dfbempty
        └── hassub/
            └── subdir/
                └── .dfbempty
        """
    )
    assert test.tree() == dedent(
        """\
        /
        ├── empty/
        │   └── file.txt
        └── hassub/
            └── subdir_moved/
                └── .dfbempty
        """
    )

    files = (os.path.relpath(f, "dst") for f in testutils.tree("dst", hidden=True))
    files = {f for f in files if not f.startswith(".dfb/")}

    assert {
        "empty/.dfbempty.19700101000001",
        "empty/.dfbempty.19700101000003D",
        "empty/file.19700101000003.txt",
        "hassub/subdir/.dfbempty.19700101000001",
        "hassub/subdir/.dfbempty.19700101000003D",
        "hassub/subdir_moved/.dfbempty.19700101000003",
    } == files

    # Now the empty dir markers shoudl go away when not tracking
    test.backup("-v", "--override", "empty_directory_markers = False", offset=5)
    assert os.path.exists("dst/hassub/subdir_moved/.dfbempty.19700101000005D")

    # and back
    res = test.backup("-v", offset=7)
    assert os.path.exists("dst/hassub/subdir_moved/.dfbempty.19700101000007")

    # restores
    test.call("restore", "restore_curr")
    assert {
        "restore_curr/empty/file.txt",
        "restore_curr/hassub/subdir_moved/.dfbempty",
    } == set(testutils.tree("restore_curr/", hidden=True))

    test.call("restore", "restore2", "--before", "u2")
    assert {
        "restore2/empty/.dfbempty",
        "restore2/hassub/subdir/.dfbempty",
    } == set(testutils.tree("restore2/", hidden=True))


def test_refresh_no_snapshots():
    test = testutils.Tester(name="refresh_no_snapshots")
    test.write_config()

    test.write_pre("src/A.txt", "A")
    test.write_pre("src/B.txt", "B")
    test.backup("-v", offset=1)

    test.write_pre("src/A.txt", "AA")
    test.backup("-v", offset=3)

    test.write_pre("src/B.txt", "BB")
    test.backup("-v", offset=5)

    shutil.rmtree("dst/.dfb/")
    test.call("refresh", "-vv")  # This used to throw errors on pulling the snapshot

    log = test.logs[-1][0]
    assert "Unable to load snapshots from remote" in log


if __name__ == "__main__":
    test_main("reference")
    #     test_main("copy")
    #     test_shell()
    #     test_log_upload()
    #     test_dst_compare_and_dst_renames(False)
    #     test_restore_error()
    #     for mode in ["size", "mtime", "hash":
    #         test_false_negs_compare(mode)
    #     test_missing_ref()
    #     test_override()
    #     test_subdirs()
    #     test_subdir_w_empty()
    #     test_symlinks("link")
    #     test_symlinks("link-webdav")
    #     test_symlinks('copy')
    #     test_symlinks("skip")
    #     test_symlinks_in_union()
    #     test_snapshots()
    #     test_missing_hashes()
    #     test_metadata(True)
    #     test_metadata(False)
    #     test_dump()
    #     test_auto()
    #     test_min_size()
    #     test_push_snapshots()
    #     test_empty_dirs()
    #     test_refresh_no_snapshots()
    print("=" * 50)
    print(" All Passed ".center(50, "="))
    print("=" * 50)

    #     #test_shell_scripts('reference')
    #     #test_shell_scripts('copy')



"""
Main backup object. This will list the source (and reset/list the dest if --refresh)
then compare.
"""

import sys, os
import time
import json
import tempfile
import subprocess
import shlex
import atexit
import shutil
import queue
import logging
import math
import gzip as gz
from collections import defaultdict
from textwrap import dedent
from threading import Thread
from functools import partial

from . import LOCK, MIN_RCLONE
from .dstdb import DFBDST, apath2rpath
from .rclonerc import IGNORED_FILE_DATA, rcpathjoin
from .threadmapper import ReturnThread, thread_map_unordered as tmap
from .utils import (
    star,
    human_readable_bytes,
    shell_runner,
    time_format,
    listify,
    smart_open,
)

# For testing only
from . import _FAIL

logger = logging.getLogger(__name__)

DFB_EMPTY = ".dfbempty"


class NoCommonHashError(ValueError):
    pass


class LinkError(ValueError, OSError):
    pass


class Backup:
    def __init__(self, config):
        self.t0 = time.time()
        self.config = config
        self.errcount = 0

    def run(self):
        config = self.config
        cliconfig = config.cliconfig

        self.call_shell(mode="pre")

        self.src_rclone = config._config["src_rclone"]

        ver = self.config.rc.call("core/version")
        logger.info("rclone version: " + ".".join(str(i) for i in ver["decomposed"]))
        for k, v in ver.items():
            logger.debug(f"   {k}: {v}")

        ver = tuple(int(c) for c in self.src_rclone.version_dict["decomposed"])
        if ver < MIN_RCLONE:
            raise ValueError(
                "Unsupported rclone version. "
                f"Must use {'.'.join(f'{i}' for i in MIN_RCLONE)} or newer"
            )

        self.dstdb = DFBDST(config)

        # Step 1: List Files locally and maybe on remote
        self.list_files()  # self.src_files, self.dst_files

        # Step 2: Compare
        self.compare()  # sets new, modified, deleted, and update_dstdb

        # update dstdb for files that match on dst_compare so that they can
        # use [src]compare next time.
        self.dstdb.replace_many(self.update_dstdb)

        # Step 3: Move Tracking
        self.track_moves()  # updates new, deleted and adds moves (original_dfile,moved_sfile)

        self.action_summary()

        if cliconfig.dry_run:
            logger.info("DRY-RUN. Exit")
            return
        elif cliconfig.interactive:
            r = input("Do you want to continue? [Y]/N:")
            if r.lower().startswith("n"):
                return

        self.dump = []

        # Step 4: Transfers. If --dump, will not act but will populate self.dump
        self.transfer()
        self.reference() if config.rename_method == "reference" else self.move_by_copy()
        self.delete()

        if file := cliconfig.dump:
            try:
                fp = smart_open(file, "wt") if file != "-" else sys.stdout
                for item in self.dump:
                    print(
                        json.dumps(item, ensure_ascii=False, separators=(",", ":")),
                        file=fp,
                        flush=True,
                    )
            finally:
                if file != "-":
                    fp.close()
                    logger.info(f"Written to {file!r}")
            return

        stats = self.run_stats()
        logger.info("-----")
        for line in stats.split("\n"):
            logger.info(line)
        logger.info("-----")

        if not cliconfig.dry_run:
            self.dstdb.push_snapshots()

        self.call_shell(mode="post", stats=stats)

        if not cliconfig.dry_run:
            self.upload_logs()

    def list_files(self, stats=None):
        """List the source and refresh dest if needed"""
        config = self.config
        kwargs = dict(stats=stats or self.config.stats)

        if self.config.cliconfig.refresh:
            sthread = ReturnThread(target=self.list_src, kwargs=kwargs).start()
            dthread = ReturnThread(
                target=self.dstdb.reset,
                kwargs=kwargs | {"use_snapshots": config.cliconfig.use_snapshots},
            ).start()
            source_files = sthread.join()
            dthread.join()

            self._proc_dst_files()
        else:
            # when we don't have to refresh, do this before listing the source just
            # so the user has some idea of how many files to possibly expect
            self._proc_dst_files()

            logger.info("Listing source")
            source_files = self.list_src(**kwargs)

        self.src_files = {file["apath"]: file for file in source_files}

        logger.info(f"Found {len(self.src_files)} source Files")

    def _proc_dst_files(self):
        d = self.dstdb.snapshot(path=self.config.cliconfig.subdir)
        d = (self.dstdb.fullrow2dict(row) for row in d)
        self.dst_files = {file["apath"]: file for file in d}
        logger.info(f"Backup contains {len(self.dst_files)} current files")

    def list_src(self, stats=None):
        config = self.config

        fsroot = config.rc.features(config.src).get("Root", "")
        if fsroot and not os.path.exists(fsroot):
            fsroot = ""

        flags = []

        compute_hashes = (
            config.get_hashes or config.compare == "hash" or config.renames == "hash"
        )

        modtime = (
            config.get_modtime
            or config.compare == "mtime"
            or config.dst_compare == "mtime"
            or config.renames == "mtime"
        )

        logger.debug(f"{compute_hashes = }, {modtime = }")

        hash_flags = []
        if compute_hashes:
            hash_flags.append("--hash")
            for htype in listify(config.hash_type):
                hash_flags.extend(["--hash-type", htype])

        subdir = config.cliconfig.subdir or ""  # Make it empty instead of None
        if subdir:
            msg = f"subdir {subdir!r} specified. Filters may break!"
            logger.warning(msg)

        rcfiles = config.src_rclone.listremote(
            filter_flags=config.filter_flags,
            # fast_list=... # Would be in rclone_flags. Already set
            mimetype=False,
            modtime=modtime,
            metadata=config.metadata,
            # only="files",
            epoch_time=True,
            flags=flags + hash_flags,
            subdir=subdir,
        )

        files = []
        dirs = set()
        parents = set()

        t0 = time.time()
        c = 0

        for item in rcfiles:
            if item["IsDir"]:
                dirs.add(os.path.join(subdir, item["Path"]))

                # could be nested w/o files so add the parent just in case
                pdir = os.path.join(subdir, os.path.dirname(item["Path"]))
                parents.add(pdir.removesuffix("/"))

                continue
            else:
                file = item

            c += 1
            new = {
                "apath": os.path.join(subdir, file.pop("Path")),
                "size": file.pop("Size"),
                "mtime": file.pop("ModTime", None),
            }

            if hashes := file.pop("Hashes", None):
                new["checksum"] = hashes

            for k, v in file.items():
                if k in IGNORED_FILE_DATA:
                    continue
                new[k] = v

            files.append(new)
            parents.add(os.path.dirname(new["apath"]))

            if stats and (time.time() - t0) >= stats:  # TODO TEST
                logger.info(f"Source Listing Status: {c} items")
                t0 = time.time()

        # Testing
        if "missing_hashes" in _FAIL:
            for file in files:
                file.pop("checksum", None)
        # end testing

        empty = dirs - parents
        if config.empty_directory_markers:
            for edir in empty:
                new = {
                    "apath": os.path.join(edir, DFB_EMPTY),
                    "mtime": -12345,
                    "size": 0,
                }
                files.append(new)

        logger.debug(f"Listed {len(files)} files")
        return files

    def compare(self):
        config = self.config

        self.new = []
        self.modified = []
        self.deleted = list(set(self.dst_files) - set(self.src_files))
        self.update_dstdb = []

        for apath, sfile in self.src_files.items():
            try:
                dfile = self.dst_files[apath]
            except KeyError:
                self.new.append(apath)
                continue

            if os.path.basename(apath) == DFB_EMPTY:
                compare = True  # Always compare empties to true regardless of attribs
            else:
                compare = self.file_compare(sfile, dfile)

            if not compare:
                self.modified.append(apath)
                continue

            # They match! But see if we need to update the dstdb with the better
            # information at source. This enables things like using mtime for
            # source-to-source but not for source-to-dest
            if dfile["dstinfo"]:
                logger.debug(f"Updating {apath!r} with src info")
                new = dfile.copy()
                new.update(sfile)
                new["dstinfo"] = 0
                self.update_dstdb.append(new)

    def file_compare(self, sfile, dfile, attrib=None):
        config = self.config

        attrib = attrib or (
            config.dst_compare if dfile["dstinfo"] else self.config.compare
        )

        msg = [f"Compare {sfile['apath']!r} with {attrib = }."]
        try:
            s = sfile.get("size", "src_missing_size")
            d = dfile.get("size", "dst_missing_size")
            if s != d:
                msg.append(f"Mismatch sizes. src: {s}, dst: {d}.")
                return False

            if attrib == "mtime":
                s = sfile.get("mtime", "src_missing_mtime")
                d = dfile.get("mtime", "dst_missing_mtime")
                try:
                    c = abs(s - d) < config.dt
                except TypeError:
                    c = False
                if not c:
                    msg.append(f"Mismatch mtime. src: {s}, dst: {d}.")
                    return False

            if attrib == "hash":
                scheck = sfile.get("checksum", {}) or {}  # Nones to empty dict
                dcheck = dfile.get("checksum", {}) or {}

                # This is a different case than no shared hashes. This happens when a remote
                # doesn't return the hashes. Like rclone itself [1,2]. Ideally, we would
                # have a settable fallback such as ModTime but this happens after listing
                # and we don't want to have to list all ModTimes on the off chance of a
                # fallback
                #     [1] https://rclone.org/flags/
                #         "-c, --checksum  Skip based on checksum (if available) & size,
                #         not mod-time & size"
                #     [2] https://forum.rclone.org/t/behavior-of-rclone-when-checksum-
                #         but-checksum-is-missing-is-undocumented-and-unexpected/39231/3
                #
                if (not scheck or not dcheck) and not config.error_on_missing_hash:
                    logger.warning(f"Missing hashes on source and/or dest")
                    logger.warning(f"  src: {sfile['apath']!r}")
                    logger.warning(f"  dst: {dfile['rpath']!r}")
                    logger.warning(f"Reverting to 'size' only")

                shared_hashes = set(scheck).intersection(set(dcheck))
                if not shared_hashes and config.error_on_missing_hash:
                    m = "Non compatible (or non existent) hashes. Change attributes"
                    logger.info(m)
                    msg.append(m)
                    msg.append(f"source = {list(scheck)}, dest = {list(dcheck)}")
                    raise NoCommonHashError(m)

                for hashname in shared_hashes:
                    if scheck[hashname] != dcheck[hashname]:
                        msg.append(f"Checksum {hashname} does not match")
                        return False

            msg.append("MATCH")
            return True
        finally:
            msg = " ".join(msg)
            logger.debug(msg)

    def track_moves(self):
        renames = self.config.renames
        dst_renames = self.config.dst_renames

        self.moves = []

        if not self.deleted or not self.new:
            logger.info("No new *and* deleted files. No rename tracking")
            return

        # The algorithm for this is pretty simple. When a file is
        # renamed, it looks like the old file is deleted and the
        # new file is created. So the candidates are pretty simple.
        #
        # Since size must *always* match, we make a dictionary by sizes
        # to reduce the pool

        del_by_size = defaultdict(list)
        for apath in self.deleted:
            dfile = self.dst_files[apath]
            del_by_size[dfile["size"]].append(dfile)

        for apath in self.new:
            if os.path.basename(apath) == DFB_EMPTY:
                continue

            sfile = self.src_files[apath]

            if (
                self.config.min_rename_size
                and sfile["size"] <= self.config.min_rename_size
            ):
                logger.debug(
                    f"Skipped rename track on {sfile['apath']!r}. "
                    f"size = {sfile['size']} <= min_rename_size = "
                    f"{self.config.min_rename_size}"
                )
                continue

            dfiles0 = del_by_size[sfile["size"]]  # list of candidate paths
            dfiles = []
            for dfile in dfiles0:
                # Note that in the config dst_renames is already set to the correct
                # values if it was None
                attrib = dst_renames if dfile.get("dstinfo", 0) else renames
                if not attrib:
                    continue

                if self.file_compare(sfile, dfile, attrib=attrib):
                    dfiles.append(dfile)

            if len(dfiles) == 1:
                self.moves.append((dfiles[0], sfile))  # dfile,moved sfile
            elif not dfiles:
                logger.debug(f"no moves for deleted file {apath!r}")
                continue
            else:
                logger.info(f"Too many matches for {apath!r}. Not moving")

        # Now we need to remove the moves from new and delete
        undelete = set()
        unnew = set()
        for dfile, moved_sfile in self.moves:
            undelete.add(dfile["apath"])
            unnew.add(moved_sfile["apath"])

        self.new[:] = list(set(self.new) - unnew)
        # DO NOT UNDELETE!!! We still want them to be "deleted" with a delete marker
        # NO: self.deleted[:] = list(set(self.deleted) - undelete)

    def transfer(self):
        config = self.config
        # dst_rclone = self.config.dst_rclone

        # The upload pipeline is done in a functional programing(esque) fashion
        # to better enable concurrency.

        comb = self.new + self.modified

        N = len(comb)
        totsize = sum(self.src_files[f]["size"] for f in comb)

        def _apath2file(apath):
            ts = self.config.now.ts

            file = self.src_files[apath].copy()
            file["rpath"] = rpath = apath2rpath(file["apath"], ts)
            file["timestamp"] = ts
            file["dstinfo"] = False  # Since this is coming from the source
            return file

        apaths = iter(comb)
        files = map(_apath2file, apaths)

        if config.cliconfig.dump:
            self.dump.extend(files)
            return

        rc = self.config.rc
        rc.start()

        def _transfer(file):
            try:
                sfile = self.config.src, file["apath"]
                dfile = self.config.dst, file["rpath"]

                if os.path.basename(file["apath"]) == DFB_EMPTY:
                    rc.write(dfile, b"")  # Empty file. mtime doesn't matter
                    logger.info(f"Uploading empty dir marker {file['rpath']!r}")
                    return file

                msg = f"Uploading {file['apath']!r} to {file['rpath']!r}"
                logger.info(msg)

                meta = self.config.metadata
                if sfile[1].endswith(".rclonelink"):
                    meta = False

                rc.copyfile(
                    src=sfile,
                    dst=dfile,
                    _config={
                        "NoCheckDest": True,
                        "metadata": meta,
                    },
                )
                return file
            except Exception as EE:
                logger.error(f"Upload Error: {file['apath']!r}. {EE}")
                with LOCK:
                    self.errcount += 1

        stats = StatsThread(self.config, N, totsize, daemon=True).start()

        files = tmap(_transfer, files, Nt=config.concurrency)
        files = filter(bool, files)
        # We could theoretically do an insert_many but that could lock the DB and/or
        # require we accumulate. Instead, let it go right to insert which closes the DB
        # each time. This is trivial compared to upload times. Plus, we want an upload
        # to be recorded in the DB right away
        files = map(self.dstdb.insert, files)

        # Make them work
        for file in files:
            stats += 1

        stats.join()

    def reference(self):
        config = self.config

        # The upload pipeline is done in a functional(esque) fashion
        # to better enable concurrency.
        moves = iter(self.moves)

        # Moves are already paired as original_dfile,moved_sfile
        def _build_new_file(original_dfile, moved_sfile):
            ts = self.config.now.ts

            new = original_dfile.copy()
            new.update(moved_sfile)

            new["original"] = original_dfile["apath"]
            new["isref"] = True
            new["rpath"] = original_dfile["rpath"]
            new["ref_rpath"] = apath2rpath(new["apath"], ts, flag="R")
            new["timestamp"] = ts
            new["dstinfo"] = False  # Since this is coming from the source
            return new

        files = map(star(_build_new_file), moves)

        if config.cliconfig.dump:
            self.dump.extend(files)
            return

        rc = self.config.rc
        rc.start()

        def _upload_ref(file):
            original = file["original"]
            ref_rpath = file["ref_rpath"]
            rpath = file["rpath"]

            ref = {
                "ver": 2,
                "rel": os.path.relpath(rpath, os.path.dirname(ref_rpath)),
            }
            reftxt = json.dumps(ref)
            try:
                logger.info(
                    f"Moving {original!r} to "
                    f"{file['apath']!r} with "
                    f"{ref_rpath!r}."
                )
                rc.write(
                    (config.dst, ref_rpath),
                    reftxt,
                )
                return file
            except Exception as EE:
                logger.error(f"Reference Error: {file['apath']!r}. {EE}")
                with LOCK:
                    self.errcount += 1

        files = tmap(_upload_ref, files, Nt=config.concurrency)
        files = filter(bool, files)
        files = map(self.dstdb.insert, files)

        # Make them work
        for file in files:
            pass

    def move_by_copy(self):
        config = self.config

        # The upload pipeline is done in a functional programing(esque) fashion
        # to better enable concurrency.
        moves = iter(self.moves)

        def _build_copiedfile(original_dfile, moved_sfile):
            ts = self.config.now.ts

            new = original_dfile.copy()
            new.update(moved_sfile)

            new["original"] = original_dfile["apath"]
            new["rpath"] = apath2rpath(new["apath"], ts)
            new["source_rpath"] = original_dfile["rpath"]
            new["timestamp"] = ts
            new["isref"] = False
            new["dstinfo"] = False  # Since this is coming from the source
            return new

        files = map(star(_build_copiedfile), moves)

        if config.cliconfig.dump:
            self.dump.extend(files)
            return

        rc = self.config.rc
        rc.start()

        def _copy(file):
            try:
                msg = f'"Moving" {file["original"]!r} to {file["apath"]!r} via copy'

                sfile = rcpathjoin(self.config.dst, file["source_rpath"])
                dfile = rcpathjoin(self.config.dst, file["rpath"])

                logger.info(msg)

                rc.copyfile(
                    src=sfile,
                    dst=dfile,
                    _config={
                        "NoCheckDest": True,
                        "metadata": self.config.metadata,
                    },
                )
                return file
            except Exception as EE:
                logger.error(f"Copy Error: {file['apath']!r}. {EE}")
                with LOCK:
                    self.errcount += 1

        files = tmap(_copy, files, Nt=config.concurrency)
        files = filter(bool, files)
        files = map(self.dstdb.insert, files)

        # Make them work
        for file in files:
            pass

    def delete(self):
        config = self.config
        # dst_rclone = self.config.dst_rclone

        # The upload pipeline is done in a functional programing(esque) fashion
        # to better enable concurrency.
        apaths = iter(self.deleted)

        def _apath2file(apath):
            ts = self.config.now.ts

            file = self.dst_files[apath].copy()
            file["rpath"] = rpath = apath2rpath(file["apath"], ts, flag="D")
            file["timestamp"] = ts
            file["dstinfo"] = False  # Since this is coming from the source
            file["size"] = -1
            return file

        files = map(_apath2file, apaths)

        if config.cliconfig.dump:
            self.dump.extend(files)
            return

        rc = self.config.rc
        rc.start()

        def _delete(file):
            dfile = file["rpath"]
            try:
                logger.info(f"Deleting {file['apath']!r} with {dfile!r}.")
                rc.write((config.dst, dfile), b"DEL")
                return file
            except Exception as EE:
                logger.error(f"Delete Error: {file['apath']!r}. {EE}")
                with LOCK:
                    self.errcount += 1

        files = tmap(_delete, files, Nt=config.concurrency)
        files = filter(bool, files)
        files = map(self.dstdb.insert, files)

        # Make them work
        for file in files:
            pass

    def action_summary(self):
        self.action_summary_text = []

        _p = logger.debug
        if self.config.cliconfig.dry_run or self.config.cliconfig.interactive:
            _p = logger.info

        m = f"New: {self.summary(self.new)}"
        self.action_summary_text.append(m)
        logger.info(m)
        for file in self.new:
            _p(f"   {file!r}")

        m = f"Modified: {self.summary(self.modified)}"
        self.action_summary_text.append(m)
        logger.info(m)
        for file in self.modified:
            _p(f"   {file!r}")

        m = f"Deleted: {self.summary(self.deleted,src=False)}"
        self.action_summary_text.append(m)
        logger.info(m)
        for file in self.deleted:
            _p(f"   {file!r}")

        m = f"Moves: {self.summary([f[0]['apath'] for f in self.moves],src=False)}"
        self.action_summary_text.append(m)
        logger.info(m)
        for file in self.moves:
            _p(f"   {file[0]['apath']!r} --> {file[1]['apath']!r}")

    def summary(self, files, src=True):
        flist = self.src_files if src else self.dst_files
        size = sum(flist[file]["size"] for file in files)
        num, units = human_readable_bytes(size)
        s = "s" if len(files) != 1 else ""
        return f"{len(files)} file{s} ({num:0.2f} {units})"

    def run_stats(self):
        stats = []

        now = self.config.now
        stats.append(
            f"Timestamp: {now.dt} "
            f"({now.obj.astimezone().strftime('%Y-%m-%d %H:%M:%S%z')})"
        )

        stats.append(f"Errors: {self.errcount}")
        select = """
            SUM(CASE 
                WHEN (size >= 0 AND (isref IS NULL OR isref = 0) )
                THEN size ELSE 0 END) 
            AS totsize,
            COUNT(size) as num
            """
        cur = self.dstdb.snapshot(select=select).fetchone()
        tot = cur["totsize"] or 0
        num, units = human_readable_bytes(tot)
        s = "s" if cur["num"] != 1 else ""
        stats.append(f"Current {cur['num']} file{s} ({num:0.2f} {units})")

        tot = self.dstdb.db().execute(f"SELECT {select} FROM items").fetchone()
        num, units = human_readable_bytes(tot["totsize"])
        s = "s" if tot["num"] != 1 else ""
        stats.append(f"Total {tot['num']} file{s} ({num:0.2f} {units})")

        stats.extend(self.action_summary_text)
        stats.append(f"Elapsed Time (approx): {time_format(time.time() - self.t0)}")
        return "\n".join(stats)

    def call_shell(self, *, mode, stats=""):
        dry = self.config.cliconfig.dry_run

        if mode == "pre":
            cmds = self.config.pre_shell
        elif mode == "post":
            cmds = self.config.post_shell

        if not cmds:
            logger.debug(f"No cmds for {mode = }")
            return

        env = {
            "CONFIGDIR": self.config._config["__dir__"],
            "STATS": stats,
            "ERRS": str(self.errcount),
        }

        returncode = shell_runner(cmds, dry=dry, env=env, prefix=f"{mode}.shell")

        if returncode and self.config.stop_on_shell_error:
            raise subprocess.CalledProcessError(returncode, cmds)

    def upload_logs(self):
        config = self.config

        if not config.logfile.exists():
            return

        name = f"{self.config.now.dt}Z.log"
        log_dests = [rcpathjoin(l, name) for l in listify(config.log_dest)]

        log_dests.append((config.dst, f".dfb/logs/{name}"))

        if not log_dests:
            logger.debug("no log destinations")
            return

        # Need to copy the log file since it may change in the process of the upload
        # from the calls itself
        log_copy = config.logfile.with_stem("log_copy")
        shutil.copy2(config.logfile, log_copy)

        for log_dest in log_dests:
            dtxt = rcpathjoin(*listify(log_dest))
            logger.info(f"Uploading log to {dtxt!r}")
            try:
                config.rc.copyfile(
                    src=log_copy,
                    dst=log_dest,
                )
            except Exception as e:
                logger.error(f"Failed: {e}")


class StatsThread(Thread):
    def __init__(self, config, N, totsize, *args, **kwargs):
        self.config = config
        self.N = N
        self.totsize = totsize
        self.fcount = 0

        # Rather than a while loop with a time.sleep and a conditional,
        # instead use a queue with a timeout. This means we can put something
        # in the queue to kill it right away. (Can this use events instead?)
        self.stop = queue.Queue()

        super().__init__(*args, **kwargs)

    def start(self, *args, **kwargs):
        super().start(*args, **kwargs)
        return self

    def increment(self, n=1):
        with LOCK:
            self.fcount += n
        return self

    __iadd__ = increment  # += n

    def run(self):
        inf = float("inf")

        totnum, totunits = human_readable_bytes(self.totsize)

        self.config.rc.call("core/stats-reset")
        self.config.rc.call("core/stats")  # sets to 0
        while True:
            try:
                stop = self.stop.get(block=True, timeout=self.config.stats)
                if stop:
                    break
            except queue.Empty:
                pass

            stats = self.config.rc.call("core/stats")
            msg = [f"STATS:"]

            dt = time_format(stats.get("elapsedTime", inf))
            msg.append(f"{dt:5s};")

            msg.append(f"xfer {len(stats.get('transferring',''))};")

            bytesnum, bytesunits = human_readable_bytes(stats.get("bytes", 0))
            msg.append(
                f"{self.fcount}/{self.N} "  # stats['totalTransfers'] includes active so use self.fcount
                f"({bytesnum:6.2f} {bytesunits} / {totnum:<6.2f} {totunits});"
            )

            # use these since they are weighted averages but need to account for each
            # transfer
            speed = sum(i.get("speedAvg", 0) for i in stats.get("transferring", []))
            # This is a heuristic approach. Prefer the weighted avg speed but
            # it can drop to zero. This doesn't have to be perfect.
            rel = speed / (stats.get("speed", 0) + 1e-5 * (speed + 1))  # [0,inf)
            frac = 2 / (1 + math.exp(-10 * rel)) - 1
            sp = frac * speed + (1 - frac) * stats.get("speed", 0)

            speednum, speedunits = human_readable_bytes(sp)
            msg.append(f"{speednum:5.2f} {speedunits}/s;")

            eta = round((self.totsize - stats.get("bytes", 0)) / sp)
            msg.append(f"ETA: {time_format(eta)}")

            logger.info(" ".join(msg))

    def join(self, *a, **k):
        self.stop.put(True)
        super().join(*a, **k)
        logger.debug("Joined stats thread")



#!/usr/bin/env python
import sys, shutil
import subprocess, os, re
from pathlib import Path

# Read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "readme.md").read_text()

# This shouldn't be needed since I have python_requires set but just in case:
if sys.version_info < (3, 9):
    raise ValueError("Must use python >= 3.9")
_r = repr
import dfb


# Extract the version from the module file
def get_version():
    version_file = this_directory / "dfb" / "__init__.py"
    with open(version_file, "r") as f:
        for line in f:
            match = re.match(r"^__version__ = ['\"]([^'\"]*)['\"]", line)
            if match:
                return match.group(1)
    raise RuntimeError("Version not found in dfb/__init__.py")


dfb_version = get_version()

from setuptools import setup

# This will set the __git_version__ in __init__.py if possible then change it back after
# install.


pwd = os.path.abspath(os.path.dirname(__file__))
try:
    version = (
        subprocess.check_output(["git", "log", "-1", "--format=%h"], cwd=pwd)
        .decode("utf8")
        .strip()
    )

    origin = subprocess.check_output(["git", "remote", "-v"], cwd=pwd).decode()
    origin = re.search(r"origin\s+?(\S.*?)\s+?\(fetch\)", origin)
    if origin:
        origin = origin.group(1)

    git = {"version": version, "origin": origin}

except:
    git = None

try:
    # Hack to replace the version. Make a copy, replace the text, then move the copy
    # back again.
    shutil.move("dfb/__init__.py", "_tmp.py")
    with open("_tmp.py") as fp:
        init = fp.read()
    init = init.replace("__git_version__ = None", f"__git_version__ = {git!r}")
    with open("dfb/__init__.py", "wt") as fp:
        fp.write(init)

    setup(
        name="dfb",
        packages=["dfb", "dfbmount", "dfblink"],
        long_description=long_description,
        install_requires=["requests"],
        entry_points={
            "console_scripts": [
                "dfb=dfb.cli:cli",
                "dfbshebanged=dfb.cli:clishebang",
                "dfb-mount=dfbmount.mount:cli",
                "dfb-link=dfblink.cli:cli",
            ],
        },
        version=dfb.__version__,  # Do not include git as per PEP440
        description="Dated File Backup",
        long_description_content_type="text/markdown",
        url="https://github.com/Jwink3101/dfb/",
        author="Justin Winokur",
        author_email="Jwink3101@users.noreply.github.com",
        license="MIT",
        python_requires=">=3.9",
    )
finally:
    shutil.move("_tmp.py", "dfb/__init__.py")



import logging
import os
from threading import Thread
from queue import Queue


logger = logging.getLogger(__name__)


class _KILL:
    pass


class ReturnThread(Thread):
    """
    Like a regular thread except when you `join`, it returns the function
    result. And .start() will return itself to enable cleaner code.

        >>> mythread = ReturnThread(...).start() # instantiate and start

    Note that target is a required keyword argument.
    """

    def __init__(self, *, target, **kwargs):
        self.target = target
        super().__init__(target=self._target, **kwargs)
        self._res = None

    def start(self, *args, **kwargs):
        super().start(*args, **kwargs)
        return self

    def _target(self, *args, **kwargs):
        self._res = self.target(*args, **kwargs)

    def join(self, *args, **kwargs):
        super().join(*args, **kwargs)
        return self._res


# I like my thread_map_unordered more since it provides better control of the buffers
# and I am pretty sure it works just fine. I have an alternative in comments that I used
# when I was getting a deadlock. It turned out to be an sqlite3 one (due to an
# executemany and a select further up the chain). I will keep it in comments here in case
# I need to go back to it.

# import multiprocessing.dummy as mpd
# def thread_map_unordered(fun,seq,Nt=None,**_):
#     with mpd.Pool(processes=Nt) as pool:
#         yield from pool.imap_unordered(fun,seq)


def thread_map_unordered(
    fun,
    seq,
    Nt=None,
    *,
    Nin_buffer=1,
    Nout_buffer=None,
    raise_exceptions=True,
):
    """
    Simple parallel mapping function.

    Inputs:
    -------
    fun
        Function to map

    seq
        Sequence

    Nt [None]
        Number of threads. Defaults to os.cpu_count()

    Nin_buffer [1]
        Number of input items to pull at a time. Set to -1 to be infinite (will exhaust
        the iterator as fast as possible). A value of 1 will most closely resemble
        a regular map() call

    Nout_buffer [None]
        Amount to buffer out. It is good to make this a bit larger than the input
        so that the threads can keep working. If None, will be Nt meaning 2*Nt can be
        cached

    raise_exceptions [True]
        If True (default), will raise exceptions.

        If false, it will *return* the exception object Note that an additional attribute
        'seq_index' will be set with the offending index and 'thread_map_unordered_exception' will be True.
        This can also be used, if needed, to know whether the exception came from this function or 'fun'
        in the off chance that an exception is a valid return from 'fun'.


    Why:
    ----
    This function seems uneccesary since there is ThreadPoolExecutor and multiprocessing.dummy.Pool.
    Both of those functions are perfectly fine but this function has two defining differences:

    (1) The ability to control how quickly the input sequence is exhausted and results yielded.
        The other tools will pull all items as fast as possible which can be very memory
        intensive. The default value of one means that it will iterate just as fast as it can but
        never faster. This could bottleneck upstream but that is a good thing for memory usage and control

    (2) Can automatically wrap exceptions without an additional code. See options above
    """
    # Use to know if an exceptoion was raised here and ONLY here. Will be replaced
    thread_map_unordered_exception = os.urandom(5)
    kill = _KILL()

    Nt = Nt or os.cpu_count()

    # Limit the input queue so as to not pull the input iterator too quickly
    qin = Queue(maxsize=Nin_buffer)
    qout = Queue(maxsize=Nout_buffer or Nt)

    def _adder():
        for ii, item in enumerate(seq):
            qin.put((ii, item))
        for _ in range(Nt):
            qin.put((-1, kill))

    adder_thread = Thread(target=_adder)
    adder_thread.start()

    def _worker():
        while True:
            ii, item = qin.get()
            if item is kill:
                qout.put(item)
                qin.task_done()
                break

            try:
                res = fun(item)
            except Exception as _res:
                res = _res  # I don't get it but this is needed
                res.seq_index = ii
                res.thread_map_unordered_exception = thread_map_unordered_exception

            qout.put(res)
            qin.task_done()

    worker_threads = [Thread(target=_worker) for _ in range(Nt)]
    for worker_thread in worker_threads:
        worker_thread.start()

    tcount = 0
    while tcount < Nt:
        res = qout.get()

        # Handle exceptions back in the main thread
        if (
            isinstance(res, Exception)
            and getattr(res, "thread_map_unordered_exception", False)
            == thread_map_unordered_exception
        ):
            res.thread_map_unordered_exception = True  # reset
            if raise_exceptions:
                raise res

        if res is kill:
            tcount += 1
            qout.task_done()
            continue
        yield res
        qout.task_done()

    qin.join()
    qout.join()
    adder_thread.join()
    for worker_thread in worker_threads:
        worker_thread.join()



"""
Restores
"""

import os, sys
import subprocess
import shlex
import logging

from . import LOCK
from .dstdb import DFBDST
from .rclonerc import rcpathjoin, rcpathsplit
from .utils import human_readable_bytes, star, listify, shell_header
from .threadmapper import thread_map_unordered as tmap

logger = logging.getLogger(__name__)


class SourceNotFoundError(ValueError):
    pass


class Restore:
    def __init__(self, config):
        self.config = config
        self.args = args = config.cliconfig

        # Handle @src alone or @src/. But we do not want @srcc to work.
        if args.dest == "@src":
            args.dest = config.src
        elif args.dest.startswith("@src/"):
            args.dest = rcpathjoin(config.src, args.dest[5:])  # 5: for /

        self.dstdb = DFBDST(config)

        if args.command == "restore-file":
            self.restore_file()
        else:
            self.restore_dir()

        self.summary()

        if not self.transfers:
            logger.info("No Transfers")
            return

        if self.args.dry_run:
            logger.info("DRY-RUN. Exit")
            return
        elif self.args.interactive:
            r = input("Do you want to continue? [Y]/N:")
            if r.lower().startswith("n"):
                return

        if self.args.shell_script:
            self.transfer_shell()
        else:
            self.transfer()

    def restore_dir(self):
        args = self.args

        snap = self.dstdb.snapshot(
            path=args.source,
            before=args.before,
            after=args.after,
            select="apath,rpath,size",
        )
        snap = map(self.dstdb.fullrow2dict, snap)

        # transfers: remote-source (rel to remote), final dest
        transfers = (
            (row["rpath"], (args.dest, row["apath"]), row["size"]) for row in snap
        )

        self.transfers = list(transfers)
        if not transfers:
            raise SourceNotFoundError(
                f"Could not find any files at {args.source!r} at the specified time"
            )

    def restore_file(self):
        args = self.args

        row = self.dstdb.snapshot(
            before=args.before,
            after=args.after,
            select="apath,rpath,size",
            conditions=[("apath = :file", {"file": args.source})],
        ).fetchone()

        if not row:
            raise SourceNotFoundError(
                f"Could not find {args.source!r} at the specified time"
            )

        if args.dest == "-":
            dest = "-"
        elif args.to:
            dest = args.dest
        else:
            dest = (args.dest, os.path.basename(args.source))

        self.transfers = [(row["rpath"], dest, row["size"])]

    def summary(self):
        _p = logger.debug
        if self.args.dry_run or self.args.interactive:
            _p = logger.info

        num, units = human_readable_bytes(sum(r[-1] for r in self.transfers))
        s = "s" if len(self.transfers) != 1 else ""
        logger.info(f"Restoring {len(self.transfers)} file{s} ({num:0.2f} {units})")

        for src, dst, size in self.transfers:
            num, units = human_readable_bytes(size)
            s = rcpathjoin(self.config.dst, src)
            d = rcpathjoin(*listify(dst))
            if d == "-":
                d = "<<stdout>>"
            _p(f"    {s!r} --> {d!r} ({num:0.2f} {units})")

    def transfer_shell(self):
        config = self.config
        out = [shell_header(config, cd=True)]

        cmd = [config.rclone_exe] + config.rclone_flags
        for src, dst, _ in self.transfers:
            src = rcpathjoin(self.config.dst, src)
            if dst == "-":
                nc = cmd + ["cat", rcpathjoin(self.config.dst, src)]
                out.append(shlex.join(nc))
                continue
            dst = rcpathjoin(*listify(dst))
            nc = cmd + ["copyto", src, dst]
            out.append(shlex.join(nc))

        if self.args.shell_script == "-":
            print("\n".join(out), flush=True)
        else:
            with open(self.args.shell_script, "wt") as fp:
                fp.write("\n".join(out))
            logger.info(f"Shell script written to {self.args.shell_script!r}")

    def transfer(self):
        config = self.config
        rc = config.rc
        rc.start()

        self.errcount = 0

        def _transfer_rc(src, dst):
            src0 = src
            try:
                src = self.config.dst, src  # ...confusing but should be the dest
                if dst == "-":
                    res = rc.read(src)
                    with LOCK:
                        try:
                            sys.stdout.buffer.write(res + b"\n")
                            sys.stdout.buffer.flush()
                        except AttributeError:
                            logger.info(
                                (
                                    "WARNING: Could not write to stdout buffer. "
                                    "Will try to decode. Otherwise, you should "
                                    "download to file"
                                ),
                                verbosity=0,
                            )
                            sys.stdout.write(res.decode() + "\n")
                            sys.stdout.flush()
                    return
                stxt = rcpathjoin(*listify(src))
                dtxt = rcpathjoin(*listify(dst))
                logger.info(f"Transfering {stxt!r} to {dtxt!r}.")

                meta = self.config.metadata
                if stxt.endswith(".rclonelink"):
                    meta = False

                rc.copyfile(
                    src=src,
                    dst=dst,
                    _config={
                        "NoCheckDest": self.args.no_check,
                        "metadata": meta,
                    },
                )
            except Exception as EE:
                msg = [f"ERROR: Could not restore {src0!r}."]
                msg.append(f"Error: {EE}")
                logger.error("\n".join(msg))
                with LOCK:
                    self.errcount += 1

        transfers = iter(self.transfers)
        transfers = (t[:2] for t in transfers)
        transfers = tmap(star(_transfer_rc), transfers, Nt=config.concurrency)
        for _ in transfers:
            pass

        if self.errcount:
            msg = "ERROR: At least one restore did not work."
            logger.info(msg)
            raise ValueError(msg)



import sys, os
import time
import tempfile
import uuid
import copy
import logging
import io
import hashlib, base64
from functools import partial, partialmethod
from pathlib import Path
from threading import Lock

logger = logging.getLogger(__name__)


LOCK = Lock()
_TEMPDIR = False  # Just used in testing

INF = float("inf")


class NotDFBFilter(logging.Filter):
    def __init__(self, include_serve=False):
        self.include_serve = include_serve
        super().__init__()

    def filter(self, record):
        if record.name.endswith("-rc-server") and not self.include_serve:
            return False
        return record.name.startswith("dfb.")


def init_logging(logfile, debuglogfile, verbosity):
    """
    Start logging. If _TEMPDIR is set, create a second one that always saves with
    more detail
    """
    USE_DEBUGFILE = bool(_TEMPDIR)

    levels = [logging.WARN, logging.INFO, logging.DEBUG]

    verbosity = min([len(levels), max([0, verbosity])])  # +1 for rc server
    if verbosity == len(levels):
        not_dfb_filter = NotDFBFilter(include_serve=True)
        verbosity -= 1
    else:
        not_dfb_filter = NotDFBFilter()

    level = levels[verbosity]

    formatter = logging.Formatter(
        fmt="%(asctime)s:%(levelname)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",  # "%z" removed
    )

    # Set up handlers with the level since the root_logger *may* be set lower
    file_handler = logging.FileHandler(logfile)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(level)
    file_handler.addFilter(not_dfb_filter)

    stream_handler = logging.StreamHandler(stream=sys.stderr)
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(level)
    stream_handler.addFilter(not_dfb_filter)

    root_logger = logging.getLogger()
    root_logger.handlers.clear()  # Need to clear them  for testing

    root_logger.setLevel(level=0 if USE_DEBUGFILE else level)
    root_logger.addHandler(file_handler)
    root_logger.addHandler(stream_handler)

    if USE_DEBUGFILE:
        dformatter = logging.Formatter(
            fmt=(
                "%(asctime)s.%(msecs)d:"  # https://stackoverflow.com/a/7517430/3633154
                "%(name)s:"
                # "%(module)s:"
                "%(lineno)d:"
                "%(funcName)s:"
                "%(levelname)s: "
                "%(message)s"
            ),
            datefmt="%Y%m%d%H%M%S",
        )

        dfile_handler = logging.FileHandler(debuglogfile)
        dfile_handler.setFormatter(dformatter)
        dfile_handler.setLevel(logging.DEBUG)
        dfile_handler.addFilter(NotDFBFilter(include_serve=True))

        root_logger.addHandler(dfile_handler)


def clean_config_id(config_id):
    """Cleans it up to alpha-numeric only"""
    config_id0 = config_id
    allowed = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-[]"
    config_id = "".join(c if c in allowed else "=" for c in config_id)

    if len(config_id) > 40:
        md5 = base64.urlsafe_b64encode(
            hashlib.md5(config_id.encode()).digest()
        ).decode()
        config_id = f"{config_id[:20]}.{md5[:8]}.{config_id[-20:]}"

    if config_id0 != config_id:
        logger.debug(f"config_id changed from {config_id0!r} to {config_id!r}")

    return config_id


class ConfigError(ValueError):
    pass


class Config:
    def __init__(self, configpath, tmpdir=None, verbosity=1, add_params=None):
        from . import nowfun, __version__, __git_version__

        self._config = {"_configpath": configpath, "verbosity": verbosity}
        try:
            self.configpath = Path(configpath).resolve()  # make it absolute
        except FileNotFoundError:
            raise FileNotFoundError(f"Couldn't find {configpath!r}")
        self.add_params = add_params or {}
        self.add_params["subdir"] = self.add_params.get("subdir", "")

        self.now = nowfun()

        if _TEMPDIR:  # Testing
            self.tmpdir = Path(_TEMPDIR)
        elif not tmpdir:
            self.tmpdir = Path(tempfile.TemporaryDirectory().name)
        else:
            self.tmpdir = Path(tmpdir) / f"{int(self.now.dt)}"

        self.tmpdir.mkdir(parents=True, exist_ok=True)

        # Start the logging
        self.logfile = self.tmpdir / "log.log"
        self.debuglogfile = self.tmpdir / "debug.log"
        init_logging(self.logfile, self.debuglogfile, verbosity)

        logger.info(f"DFB ({__version__})")
        if __git_version__:
            logger.info(f" {__git_version__['version']} {__git_version__['origin']}")
        logger.info(f"Now: {self.now.obj.astimezone().isoformat()}")
        logger.info(f"Backup Timestamp: {self.now.dt}Z")
        logger.info(f"config path: '{self.configpath}'")
        logger.info(f"tmpdir: {str(self.tmpdir)}")

    def _write_template(self, force=False):
        from . import __version__

        txt = TEMPLATE
        txt = txt.replace("__VERSION__", __version__)
        txt = txt.replace("__UUID4__", str(uuid.uuid4()))

        if self.configpath.exists() and not force:
            raise ValueError(
                f"Path '{self.configpath}' exists. "
                "Specify a different path, move the existing file "
                "or use --force-overwrite"
            )
        self.configpath.parent.mkdir(parents=True, exist_ok=True)
        self.configpath.write_text(txt)

        os.chmod(self.configpath, self.configpath.stat().st_mode | 0o100)

        logger.debug(f"Wrote template config to {self.configpath}")

    def parse(self, override_txt=""):
        from .rclonerc import RC

        if self.configpath is None:
            raise ValueError("Must have a config path")

        # Passed to the config file
        self._config["stats"] = 30  # Fixed. Undocumented that this can be set
        self._config["os"] = os
        self._config["Path"] = Path
        self._config["log"] = lambda x: logger.info(f"config: {x}")
        self._config["print"] = lambda x: logger.info(f"config: {x}")
        self._config["__file__"] = self.configpath
        self._config["__dir__"] = self.configpath.parent
        self._config["DELENV"] = RC.DELENV
        self._config["clean_config_id"] = clean_config_id

        self._config.update(self.add_params)

        self._hidden_keys = set(self._config)  # to be removed in repr
        self._hidden_keys.difference_update({"configpath", "verbosity"})

        junk = {}
        exec("", junk)

        exec(TEMPLATE, self._config)  # Defaults
        self._config_keys = [
            k
            for k in self._config
            if (k not in junk and k not in self._hidden_keys and not k.startswith("_"))
        ]
        self._config_keys.append("destpaths")

        # Read then change dir
        config_txt = self.configpath.read_text()
        # os.chdir(self._config["__dir__"])

        # Set the override_txt before AND after so that you can set other things
        cfg = [
            "pre,post = True,False",
            override_txt,
            config_txt,
            "pre,post = False,True",
            override_txt,
        ]
        exec("\n".join(cfg), self._config)

        for key in junk:
            self._config.pop(key, 0)

        self._validate()

        logger.info(f"ID: {self.config_id}")
        logger.debug(f"Read config {str(self.configpath)!r}")
        for k in self._config_keys:
            if k not in self._config:
                continue
            dispval = self._config[k]
            if k == "rclone_env":
                dispval = {
                    n: (k if n != "RCLONE_CONFIG_PASS" else "**REDACTED**")
                    for n, k in dispval.items()
                }
            logger.debug(f"   {k} = {dispval!r}")

        # Set these up because all methods will use them
        settings = dict(
            rclone_exe=self.rclone_exe,
            universal_flags=self.rclone_flags,
            universal_env=self.rclone_env,
            # Even if using something like Google Drive which allows duplicates,
            # we won't *ever* be overwriting files so this is okay.
            no_check_dest=True,
        )
        if self.metadata:
            settings["universal_flags"].append("--metadata")
            # self.rclone_flags.append("--metadata")

        # Set up rclone objects. For the most part, we use the RC interface
        # but there are a few things where the CLI is better or easier. These may
        # go away in future versions. The CLI interface (rclonecli.py) has many
        # "features" that are eclipsed by rc, but a few are better on CLI.
        #
        # Add these here since we will need them throughout the eval. Note that
        # rc is using the rclone server while (src/dst)_rclone is making CLI
        # calls to rclone.
        from .rclonecli import RcloneCLI
        from .rclonerc import RC

        self.rc = RC(
            rclone_exe=self.rclone_exe,
            serve_flags=self.rclone_flags
            + ["-vv"],  # always include verbose but filter later
            rclone_env=self.rclone_env,
        )

        self.src_rclone = RcloneCLI(self._config["src"], **settings)
        self.dst_rclone = RcloneCLI(self._config["dst"], **settings)
        # Monkey patch the debug
        self.src_rclone.debug = lambda x: logger.debug(f"src-rclone: {x}")
        self.dst_rclone.debug = lambda x: logger.debug(f"dst-rclone: {x}")

        # Set the db_cachedir here. Needed to wait for rclone cli objects
        if not (dbcache_dir := self._config.get("dbcache_dir", None)):
            dbcache_dir = Path(self.dst_rclone.config_paths["Cache dir"]) / "DFB"
        self._config["dbcache_dir"] = Path(dbcache_dir)
        self._config["snap_cache_dir"] = self.dbcache_dir / f"{self.config_id}.snap"

        return self

    def _validate(self):
        """
        Validate config
        """
        from .rclonerc import FILTER_FLAGS

        if self.src == "<<MUST SPECIFY>>":
            raise ConfigError("Must specify 'src'")
        if self.dst == "<<MUST SPECIFY>>":
            raise ConfigError("Must specify 'dst'")

        allowed = {
            "compare": {"mtime", "size", "hash", "auto"},
            "dst_compare": {"mtime", "size", "hash", "auto", None},
            "renames": {"size", "mtime", "hash", "auto", False, None},
            "dst_renames": {"size", "mtime", "hash", "auto", False, None},
            "rename_method": {"reference", "copy", False, None},
            "get_modtime": {True, False, "auto"},
            "get_hashes": {True, False, "auto"},
        }

        for key, values in allowed.items():
            val = self._config[key]
            if val not in values:
                msg = f"Allowed values for '{key}' are {values}. Specified {repr(val)}"
                raise ConfigError(msg)
        ff = FILTER_FLAGS.union(["--one-file-system"])
        if badflags := ff.intersection(self.rclone_flags):
            msg = f"May not have {badflags} in 'rclone_flags'. Use 'filter_flags'"
            raise ConfigError(msg)

        # These will also set to auto if the compare or renames is auto
        self._config["dst_compare"] = (
            self._config["dst_compare"] or self._config["compare"]
        )
        if self._config["dst_renames"] is None:  # explicit because could be False
            self._config["dst_renames"] = self._config["renames"]

        if not self._config.get("dst_atomic_transfer", True):
            logger.info(
                "WARNING: 'dst_atomic_transfer = False' is deprecated since rclone 1.63 "
                "handles it for non-atomic remotes"
            )

        if self._config.get("links", False):
            logger.info(
                "WARNING: 'links' is deprecated. The link setting should be specified in rclone_flags"
            )

        # Set the config_id but give preference to _uuid
        self._config["config_id"] = clean_config_id(
            self._config.get("_uuid", self._config["config_id"])
        )

        if mrs := self._config["min_rename_size"]:
            from .utils import parse_bytes

            self._config["min_rename_size"] = mrs1 = parse_bytes(mrs)
            logger.debug(f"Parsed min_rename_size {mrs!r} as {mrs1!r} bytes")

    def _set_auto(self):
        sf = self.rc.features(self.src)
        df = self.rc.features(self.dst)

        src_mtime = sf["Precision"] < 1.1e9 and not sf["Features"]["SlowModTime"]
        dst_mtime = df["Precision"] < 1.1e9 and not df["Features"]["SlowModTime"]

        if self.compare == "auto":  # src to src
            self.compare = "mtime" if src_mtime else "size"
            logger.debug(f"auto-setting 'compare' to {self.compare!r}")

        if self.dst_compare == "auto":  # src to dst
            # don't *just* do self.compare since it could be hash
            if self.compare != "size" and src_mtime and dst_mtime:
                self.dst_compare = "mtime"
            else:
                self.dst_compare = "size"

            logger.debug(f"auto-setting 'dst_compare' to {self.dst_compare!r}")

        if self.renames == "auto":  # src to src
            self.renames = "mtime" if src_mtime else False
            logger.debug(f"auto-setting 'renames' to {self.renames!r}")

        if self.dst_renames == "auto":  # src to dst
            # don't *just* do self.renames since it could be hash or False
            if self.renames != "size" and self.renames and src_mtime and dst_mtime:
                self.dst_renames = "mtime"
            else:
                self.dst_renames = False

            logger.debug(f"auto-setting 'dst_renames' to {self.dst_renames!r}")

        if self.get_modtime == "auto":
            self.get_modtime = src_mtime
            logger.debug(f"auto-setting 'get_modtime' to {self.get_modtime!r}")

        if self.get_hashes == "auto":
            # self.get_hashes = sf["Hashes"] and not sf["Features"]["SlowHash"]
            # if sf['String'].startswith('S3 '):
            #    logger.warning(
            #        "src is S3 remote. S3 doesn't *always* provide hashes without "
            #        "additional API calls. Setting get_hashes to False to be safe"
            #    )
            #    self.get_hashes = False
            # logger.debug(f"auto-setting 'get_hashes' to {self.get_hashes!r}")

            # to be safe, making this always false. Users should set this if they want it
            self.get_hashes = False
            logger.debug(f"setting 'get_hashes' to False regardless of remotes")

    def __getattr__(self, attr):
        return self._config[attr]

    def __setattr__(self, attr, value):
        if attr.startswith("_"):
            return super(Config, self).__setattr__(attr, value)

        self._config[attr] = value

    def __repr__(self):
        # Need to watch out for RCLONE_CONFIG_PASS in rclone_env
        # make a copy of the dict fixing that one but do not
        # just do a deepcopy in case the user imported modules
        cfg = copy.copy(self._config)
        cfg["rclone_env"] = cfg["rclone_env"].copy()

        if "RCLONE_CONFIG_PASS" in cfg["rclone_env"]:
            cfg["rclone_env"]["RCLONE_CONFIG_PASS"] = "**REDACTED**"

        contents = ", ".join(
            f"{k}={cfg[k]!r}" for k in self._config_keys if k in self._config
        )
        return f"Config({contents})"


TEMPLATE = r'''#!/usr/bin/env dfbshebanged
"""
DFB Config File

This configuration file is read as Python so things can be customized as
desired. With few exceptions, any missing items will go to the defaults
already specified.

rclone flags should always be a list. 
Example: `--exclude myfile` will be ['--exclude','myfile']

Defaults are sensible for a local source. Change as needed for others.

A few modules, including `os` and `Path = pathlib.Path` are already loaded along
with `logger.info()` and `logger.debug()`

Also defines:
    __file__ : Absolute path of the config file. pathlib.Path
    __dir__ : Absolute path of the config file parent. pathlib.Path
    subdir : Value of '--subdir' if specified, otherwise an empty string.

ALL LOCAL PATHS SHOULD BE ABSOLUTE
"""
##############################################
##              Basic Settings              ##
##                (required)                ##
##############################################

# Specify the source and destination rclone remotes
src = "<<MUST SPECIFY>>"
dst = "<<MUST SPECIFY>>"

##############################################
##                 Filters                  ##
##############################################

# Specify flags used for filtering. See https://rclone.org/filtering/
#
# If using flags like '--filter-from', they should be absolute paths. Example:
#   ["--filter-from", __dir__ / "myfilters.txt"]
#
# Note that when backups use --subdir, the paths specified here may be incorrect to the
# source. The variable 'subdir' is defined to assist. USE WITH CAUTION
filter_flags = []


##############################################
## Comparison and Rename-Tracking Settings  ##
##           (defaults to "auto")           ##
##############################################

# src-to-src comparison to determine changes. Auto tries to use mtime if supported.
compare = "auto"  # "size", "mtime", "hash", "auto"

# src-to-dst comparison such as after a refresh without snapshots. None uses 'compare'
dst_compare = None  # "size", "mtime", "hash", "auto", None

# Rename tracking src-to-src. False disables it. "size" is risky!
renames = "auto"  # "size", "mtime", "hash", "auto", False

# Rename tracking src-to-dst. None uses 'renames'
dst_renames = None  # "size", "mtime", "hash", "auto", False, None

# Rename files with server-side-copy or with reference file
rename_method = "reference"  # 'reference', 'copy'

# If files are less than this size, do not track the rename. Can specify with
# normal prefixes or an integer byte count (e.g. 2097152, "2 KiB", "10 MB", "15 MiB")
min_rename_size = 0

##############################################
##             rclone Settings              ##
##         (optional, intermediate)         ##
##############################################

# Flags and environment for general rclone usage.
# Examples include --config (or RCLONE_CONFIG).
# Example for config password
#   > from getpass import getpass
#   > rclone_env = {"RCLONE_CONFIG_PASS": getpass("Password: ")}
#
# Notes:
# - This IS where you should specify link flags for local such as 
#   --links, --skip-links, --copy-links
# - Not all flags are compatible
rclone_flags = []
rclone_env = {}

# Flags for refresh specifically. Example: --fast-list
dst_list_rclone_flags = []

# Executable
rclone_exe = "rclone"

##############################################
##          Intermediate Settings           ##
##                (optional)                ##
##############################################

# Number of transfers. Use remote-specific flags for additional control such as
# --s3-upload-concurrency.
concurrency = os.cpu_count()

# Tolerance on mtimes
dt = 1.0  # seconds

# Whether to always get src mtime. Can be expensive and slow on some remotes.
get_modtime = "auto"  # True, False, "auto"

# Allow missing hash. Falls back to "size"
error_on_missing_hash = False

# String or list of strings of hash types. None uses all available on a remote
hash_type = None

# Whether to always request hashes. "auto" maps to False for now
get_hashes = False  # True, False, "auto"

# Request and transfer (if possible) metadata. Metadata is also stored in the snapshot
# files in case the dst doesn't support it. However, it must be restored manually if not
# supported. Changes in file metadata will *not* force a backup again.
metadata = True

# Specify additional log destinations. These should be full rclone remote directories
# to save the log in addition to the remote. String or list of strings
log_dest = None

# Unique name for this config file. Default is just based on the src and dst. Will be
# processed and cleaned as needed
config_id = f"{src}-{dst}"

# Specify where to store the file database. Default (None) is
# `<rclone cache dir>/DFB/<config_id>.db`.
dbcache_dir = None

# If True, dfb will add an empty directory marker file, ".dfbempty", for empty 
# directories. These markers will be present in a restore but can later be deleted
empty_directory_markers = False

##############################################
##             Disable Features             ##
##     Use --override to undo. Example:     ##
##    --override "disable_prune = False"    ##
##############################################

# Pruning is the only distructive process in dfb. In order to make sure it isn't done by
# accident, setting this to True will force all pruneing to be as if called with
disable_prune = False

# While not destructive, refresh can be very slow on some remotes, especially with many
# references and/or not using snapshots
disable_refresh = False

##############################################
##       Pre- and Post-Shell Options        ##
##                (optional)                ##
##############################################

## Pre-, Post- and fail- shell commands to run
# Specify shell code to be evaluated before and/or after running dfb. Note
# these are all run from the current directory but $CONFIGDIR environment will be set.
# STDOUT and STDERR will be captured. Note that there is no validation or
# security of the inputs. These are not actually called if using dry-run.
#
# The post shell call also has the "$STATS" environment variable defined which has
# the run statistics including timing (which will be different than the final since
# the logs will not yet have been dumped)
#
# Can be specified as the following:
#
#     string : Run with shell=True Can cd as needed (including `cd $CONFIGDIR`) and
#              can be multiple lines and multiple commands.
#     list   : Will execute with shell=False in the current directory
#     dict   : Specify subprocess.Popen flags plus the keyword 'cmd'. YOU decide if
#              shell should be True or False based on 'cmd'. dfb will override settings
#              for std(out/err). Will update os.environ with any 'env' settings.
#
# If the final cmd if specified directly as a list or a list inside the dict, each
# command will be run through C-style formatting of the environ. C-Style is used to not
# break the more-common str.format (or f-string) formats that may exist.
# os.environ may be updated with `env` inside of a dict.
#
# Example:
#   post_shell = 'echo "$STATS"'
#   post_shell = ["echo","%(STATS)s"]
# will look the same
pre_shell = ""
post_shell = ""

# Specify whether or not to allow an error in the shell commands above to continue
stop_on_shell_error = False

# If, and only if, the run fails, this will get called. It will not be logged or
# have any information about the failure however the temporary log, $LOGPATH, and
# debug log, $DEBUGPATH, could be read if interested. Useful for sending alert of
# failed run. Same rules as above about being specified as a list
fail_shell = ""


#######
# This should only be changed by the user when migrating from an older config
# to a newer one. Just because the current version of dbf and the version
# below do not match, it does not mean the run won't work.
_version = "__VERSION__"
'''
